# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
SuperSocket.
"""

from select import select, error as select_error
import ctypes
import errno
import socket
import struct
import time

from scapy.config import conf
from scapy.consts import DARWIN, WINDOWS
from scapy.data import (
    MTU,
    ETH_P_IP,
    ETH_P_IPV6,
    SOL_PACKET,
    SO_TIMESTAMPNS,
)
from scapy.compat import raw
from scapy.error import warning, log_runtime
from scapy.interfaces import network_name
from scapy.packet import Packet, NoPayload
from scapy.plist import (
    PacketList,
    SndRcvList,
    _PacketIterable,
)
from scapy.utils import PcapReader, tcpdump

# Typing imports
from scapy.interfaces import _GlobInterfaceType
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

# Utils


class _SuperSocket_metaclass(type):
    desc = None   # type: Optional[str]

    def __repr__(self):
        # type: () -> str
        if self.desc is not None:
            return "<%s: %s>" % (self.__name__, self.desc)
        else:
            return "<%s>" % self.__name__


# Used to get ancillary data
PACKET_AUXDATA = 8
ETH_P_8021Q = 0x8100
TP_STATUS_VLAN_VALID = 1 << 4
TP_STATUS_VLAN_TPID_VALID = 1 << 6


class tpacket_auxdata(ctypes.Structure):
    _fields_ = [
        ("tp_status", ctypes.c_uint),
        ("tp_len", ctypes.c_uint),
        ("tp_snaplen", ctypes.c_uint),
        ("tp_mac", ctypes.c_ushort),
        ("tp_net", ctypes.c_ushort),
        ("tp_vlan_tci", ctypes.c_ushort),
        ("tp_vlan_tpid", ctypes.c_ushort),
    ]  # type: List[Tuple[str, Any]]


# SuperSocket

class SuperSocket(metaclass=_SuperSocket_metaclass):
    closed = False  # type: bool
    nonblocking_socket = False  # type: bool
    auxdata_available = False   # type: bool

    def __init__(self,
                 family=socket.AF_INET,  # type: int
                 type=socket.SOCK_STREAM,  # type: int
                 proto=0,  # type: int
                 iface=None,  # type: Optional[_GlobInterfaceType]
                 **kwargs  # type: Any
                 ):
        # type: (...) -> None
        self.ins = socket.socket(family, type, proto)  # type: socket.socket
        self.outs = self.ins  # type: Optional[socket.socket]
        self.promisc = conf.sniff_promisc
        self.iface = iface or conf.iface

    def send(self, x):
        # type: (Packet) -> int
        """Sends a `Packet` object

        :param x: `Packet` to be send
        :return: Number of bytes that have been sent
        """
        sx = raw(x)
        try:
            x.sent_time = time.time()
        except AttributeError:
            pass

        if self.outs:
            return self.outs.send(sx)
        else:
            return 0

    if WINDOWS:
        def _recv_raw(self, sock, x):
            # type: (socket.socket, int) -> Tuple[bytes, Any, Optional[float]]
            """Internal function to receive a Packet.

            :param sock: Socket object from which data are received
            :param x: Number of bytes to be received
            :return: Received bytes, address information and no timestamp
            """
            pkt, sa_ll = sock.recvfrom(x)
            return pkt, sa_ll, None
    else:
        def _recv_raw(self, sock, x):
            # type: (socket.socket, int) -> Tuple[bytes, Any, Optional[float]]
            """Internal function to receive a Packet,
            and process ancillary data.

            :param sock: Socket object from which data are received
            :param x: Number of bytes to be received
            :return: Received bytes, address information and an optional timestamp
            """
            timestamp = None
            if not self.auxdata_available:
                pkt, _, _, sa_ll = sock.recvmsg(x)
                return pkt, sa_ll, timestamp
            flags_len = socket.CMSG_LEN(4096)
            pkt, ancdata, flags, sa_ll = sock.recvmsg(x, flags_len)
            if not pkt:
                return pkt, sa_ll, timestamp
            for cmsg_lvl, cmsg_type, cmsg_data in ancdata:
                # Check available ancillary data
                if (cmsg_lvl == SOL_PACKET and cmsg_type == PACKET_AUXDATA):
                    # Parse AUXDATA
                    try:
                        auxdata = tpacket_auxdata.from_buffer_copy(cmsg_data)
                    except ValueError:
                        # Note: according to Python documentation, recvmsg()
                        #       can return a truncated message. A ValueError
                        #       exception likely indicates that Auxiliary
                        #       Data is not supported by the Linux kernel.
                        return pkt, sa_ll, timestamp
                    if auxdata.tp_vlan_tci != 0 or \
                            auxdata.tp_status & TP_STATUS_VLAN_VALID:
                        # Insert VLAN tag
                        tpid = ETH_P_8021Q
                        if auxdata.tp_status & TP_STATUS_VLAN_TPID_VALID:
                            tpid = auxdata.tp_vlan_tpid
                        tag = struct.pack(
                            "!HH",
                            tpid,
                            auxdata.tp_vlan_tci
                        )
                        pkt = pkt[:12] + tag + pkt[12:]
                elif cmsg_lvl == socket.SOL_SOCKET and \
                        cmsg_type == SO_TIMESTAMPNS:
                    length = len(cmsg_data)
                    if length == 16:  # __kernel_timespec
                        tmp = struct.unpack("ll", cmsg_data)
                    elif length == 8:  # timespec
                        tmp = struct.unpack("ii", cmsg_data)
                    else:
                        log_runtime.warning("Unknown timespec format.. ?!")
                        continue
                    timestamp = tmp[0] + tmp[1] * 1e-9
            return pkt, sa_ll, timestamp

    def recv_raw(self, x=MTU):
        # type: (int) -> Tuple[Optional[Type[Packet]], Optional[bytes], Optional[float]]  # noqa: E501
        """Returns a tuple containing (cls, pkt_data, time)


        :param x: Maximum number of bytes to be received, defaults to MTU
        :return: A tuple, consisting of a Packet type, the received data,
                 and a timestamp
        """
        return conf.raw_layer, self.ins.recv(x), None

    def recv(self, x=MTU, **kwargs):
        # type: (int, **Any) -> Optional[Packet]
        """Receive a Packet according to the `basecls` of this socket

        :param x: Maximum number of bytes to be received, defaults to MTU
        :return: The received `Packet` object, or None
        """
        cls, val, ts = self.recv_raw(x)
        if not val or not cls:
            return None
        try:
            pkt = cls(val, **kwargs)  # type: Packet
        except KeyboardInterrupt:
            raise
        except Exception:
            if conf.debug_dissector:
                from scapy.sendrecv import debug
                debug.crashed_on = (cls, val)
                raise
            pkt = conf.raw_layer(val)
        if ts:
            pkt.time = ts
        return pkt

    def fileno(self):
        # type: () -> int
        return self.ins.fileno()

    def close(self):
        # type: () -> None
        """Gracefully close this socket
        """
        if self.closed:
            return
        self.closed = True
        if getattr(self, "outs", None):
            if getattr(self, "ins", None) != self.outs:
                if self.outs and self.outs.fileno() != -1:
                    self.outs.close()
        if getattr(self, "ins", None):
            if self.ins.fileno() != -1:
                self.ins.close()

    def sr(self, *args, **kargs):
        # type: (Any, Any) -> Tuple[SndRcvList, PacketList]
        """Send and Receive multiple packets
        """
        from scapy import sendrecv
        return sendrecv.sndrcv(self, *args, **kargs)

    def sr1(self, *args, **kargs):
        # type: (Any, Any) -> Optional[Packet]
        """Send one packet and receive one answer
        """
        from scapy import sendrecv
        # if not explicitly specified by the user,
        # set threaded to False in sr1 to remove the overhead
        # for a Thread creation
        kargs.setdefault("threaded", False)
        ans = sendrecv.sndrcv(self, *args, **kargs)[0]  # type: SndRcvList
        if len(ans) > 0:
            pkt = ans[0][1]  # type: Packet
            return pkt
        else:
            return None

    def sniff(self, *args, **kargs):
        # type: (Any, Any) -> PacketList
        from scapy import sendrecv
        return sendrecv.sniff(opened_socket=self, *args, **kargs)

    def tshark(self, *args, **kargs):
        # type: (Any, Any) -> None
        from scapy import sendrecv
        sendrecv.tshark(opened_socket=self, *args, **kargs)

    # TODO: use 'scapy.ansmachine.AnsweringMachine' when typed
    def am(self,
           cls,  # type: Type[Any]
           *args,  # type: Any
           **kwargs  # type: Any
           ):
        # type: (...) -> Any
        """
        Creates an AnsweringMachine associated with this socket.

        :param cls: A subclass of AnsweringMachine to instantiate
        """
        return cls(*args, opened_socket=self, socket=self, **kwargs)

    @staticmethod
    def select(sockets, remain=conf.recv_poll_rate):
        # type: (List[SuperSocket], Optional[float]) -> List[SuperSocket]
        """This function is called during sendrecv() routine to select
        the available sockets.

        :param sockets: an array of sockets that need to be selected
        :returns: an array of sockets that were selected and
            the function to be called next to get the packets (i.g. recv)
        """
        try:
            inp, _, _ = select(sockets, [], [], remain)
        except (IOError, select_error) as exc:
            # select.error has no .errno attribute
            if not exc.args or exc.args[0] != errno.EINTR:
                raise
        return inp

    def __del__(self):
        # type: () -> None
        """Close the socket"""
        self.close()

    def __enter__(self):
        # type: () -> SuperSocket
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # type: (Optional[Type[BaseException]], Optional[BaseException], Optional[Any]) -> None  # noqa: E501
        """Close the socket"""
        self.close()


if not WINDOWS:
    class L3RawSocket(SuperSocket):
        desc = "Layer 3 using Raw sockets (PF_INET/SOCK_RAW)"

        def __init__(self,
                     type=ETH_P_IP,  # type: int
                     filter=None,  # type: Optional[str]
                     iface=None,  # type: Optional[_GlobInterfaceType]
                     promisc=None,  # type: Optional[bool]
                     nofilter=0  # type: int
                     ):
            # type: (...) -> None
            self.outs = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_RAW)  # noqa: E501
            self.outs.setsockopt(socket.SOL_IP, socket.IP_HDRINCL, 1)
            self.ins = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.htons(type))  # noqa: E501
            if iface is not None:
                iface = network_name(iface)
                self.iface = iface
                self.ins.bind((iface, type))
            else:
                self.iface = "any"
            try:
                # Receive Auxiliary Data (VLAN tags)
                self.ins.setsockopt(SOL_PACKET, PACKET_AUXDATA, 1)
                self.ins.setsockopt(
                    socket.SOL_SOCKET,
                    SO_TIMESTAMPNS,
                    1
                )
                self.auxdata_available = True
            except OSError:
                # Note: Auxiliary Data is only supported since
                #       Linux 2.6.21
                msg = "Your Linux Kernel does not support Auxiliary Data!"
                log_runtime.info(msg)

        def recv(self, x=MTU, **kwargs):
            # type: (int, **Any) -> Optional[Packet]
            data, sa_ll, ts = self._recv_raw(self.ins, x)
            if sa_ll[2] == socket.PACKET_OUTGOING:
                return None
            if sa_ll[3] in conf.l2types:
                cls = conf.l2types.num2layer[sa_ll[3]]  # type: Type[Packet]
                lvl = 2
            elif sa_ll[1] in conf.l3types:
                cls = conf.l3types.num2layer[sa_ll[1]]
                lvl = 3
            else:
                cls = conf.default_l2
                warning("Unable to guess type (interface=%s protocol=%#x family=%i). Using %s", sa_ll[0], sa_ll[1], sa_ll[3], cls.name)  # noqa: E501
                lvl = 3

            try:
                pkt = cls(data, **kwargs)
            except KeyboardInterrupt:
                raise
            except Exception:
                if conf.debug_dissector:
                    raise
                pkt = conf.raw_layer(data)

            if lvl == 2:
                pkt = pkt.payload

            if pkt is not None:
                if ts is None:
                    from scapy.arch.linux import get_last_packet_timestamp
                    ts = get_last_packet_timestamp(self.ins)
                pkt.time = ts
            return pkt

        def send(self, x):
            # type: (Packet) -> int
            try:
                sx = raw(x)
                if self.outs:
                    x.sent_time = time.time()
                    return self.outs.sendto(
                        sx,
                        (x.dst, 0)
                    )
            except AttributeError:
                raise ValueError(
                    "Missing 'dst' attribute in the first layer to be "
                    "sent using a native L3 socket ! (make sure you passed the "
                    "IP layer)"
                )
            except socket.error as msg:
                log_runtime.error(msg)
            return 0

    class L3RawSocket6(L3RawSocket):
        def __init__(self,
                     type: int = ETH_P_IPV6,
                     filter: Optional[str] = None,
                     iface: Optional[_GlobInterfaceType] = None,
                     promisc: Optional[bool] = None,
                     nofilter: bool = False) -> None:
            # NOTE: if fragmentation is needed, it will be done by the kernel (RFC 2292)  # noqa: E501
            self.outs = socket.socket(
                socket.AF_INET6,
                socket.SOCK_RAW,
                socket.IPPROTO_RAW
            )
            self.ins = socket.socket(
                socket.AF_PACKET,
                socket.SOCK_RAW,
                socket.htons(type)
            )
            self.iface = cast(_GlobInterfaceType, iface)


class SimpleSocket(SuperSocket):
    desc = "wrapper around a classic socket"
    __selectable_force_select__ = True

    def __init__(self, sock, basecls=None):
        # type: (socket.socket, Optional[Type[Packet]]) -> None
        self.ins = sock
        self.outs = sock
        if basecls is None:
            basecls = conf.raw_layer
        self.basecls = basecls

    def recv_raw(self, x=MTU):
        # type: (int) -> Tuple[Optional[Type[Packet]], Optional[bytes], Optional[float]]
        return self.basecls, self.ins.recv(x), None

    if WINDOWS:
        @staticmethod
        def select(sockets, remain=None):
            # type: (List[SuperSocket], Optional[float]) -> List[SuperSocket]
            from scapy.automaton import select_objects
            return select_objects(sockets, remain)


class StreamSocket(SimpleSocket):
    """
    Wrap a stream socket into a layer 2 SuperSocket

    :param sock: the socket to wrap
    :param basecls: the base class packet to use to dissect the packet
    """
    desc = "transforms a stream socket into a layer 2"

    def __init__(self,
                 sock,  # type: socket.socket
                 basecls=None,  # type: Optional[Type[Packet]]
                 ):
        # type: (...) -> None
        from scapy.sessions import streamcls
        self.rcvcls = streamcls(basecls or conf.raw_layer)
        self.metadata: Dict[str, Any] = {}
        self.streamsession: Dict[str, Any] = {}
        self._buf = b""
        super(StreamSocket, self).__init__(sock, basecls=basecls)

    def recv(self, x=None, **kwargs):
        # type: (Optional[int], Any) -> Optional[Packet]
        if x is None:
            x = MTU
        # Block but in PEEK mode
        data = self.ins.recv(x, socket.MSG_PEEK)
        if data == b"":
            raise EOFError
        x = len(data)
        pkt = self.rcvcls(self._buf + data, self.metadata, self.streamsession)
        if pkt is None:  # Incomplete packet.
            self._buf += self.ins.recv(x)
            return self.recv(x)
        self.metadata.clear()
        # Strip any madding
        pad = pkt.getlayer(conf.padding_layer)
        if pad is not None and pad.underlayer is not None:
            del pad.underlayer.payload
        while pad is not None and not isinstance(pad, NoPayload):
            x -= len(pad.load)
            pad = pad.payload
        # Only receive the packet length
        self.ins.recv(x)
        self._buf = b""
        return pkt


class StreamSocketPeekless(StreamSocket):
    desc = "StreamSocket that doesn't use MSG_PEEK"

    def __init__(self, sock, basecls=None):
        # type: (socket.socket, Optional[Type[Packet]]) -> None
        from scapy.sessions import TCPSession
        self.sess = TCPSession(app=True)
        super(StreamSocketPeekless, self).__init__(sock, basecls)

    # 65535, the default value of x is the maximum length of a TLS record
    def recv(self, x=None, **kwargs):
        # type: (Optional[int], **Any) -> Optional[Packet]
        if x is None:
            x = MTU
        # Block
        try:
            data = self.ins.recv(x)
        except OSError:
            raise EOFError
        try:
            pkt = self.sess.process(data, cls=self.basecls)  # type: ignore
        except struct.error:
            # Buffer underflow
            pkt = None
        if data == b"" and not pkt:
            raise EOFError
        if not pkt:
            return self.recv(x)
        return pkt

    @staticmethod
    def select(sockets, remain=None):
        # type: (List[SuperSocket], Optional[float]) -> List[SuperSocket]
        queued = [
            x
            for x in sockets
            if isinstance(x, StreamSocketPeekless) and x.sess.data
        ]
        if queued:
            return queued  # type: ignore
        return super(StreamSocketPeekless, StreamSocketPeekless).select(
            sockets,
            remain=remain,
        )


# Old name: SSLStreamSocket
SSLStreamSocket = StreamSocketPeekless


class L2ListenTcpdump(SuperSocket):
    desc = "read packets at layer 2 using tcpdump"

    def __init__(self,
                 iface=None,  # type: Optional[_GlobInterfaceType]
                 promisc=None,  # type: Optional[bool]
                 filter=None,  # type: Optional[str]
                 nofilter=False,  # type: bool
                 prog=None,  # type: Optional[str]
                 quiet=False,  # type: bool
                 *arg,  # type: Any
                 **karg  # type: Any
                 ):
        # type: (...) -> None
        self.outs = None
        args = ['-w', '-', '-s', '65535']
        self.iface = "any"
        if iface is None and (WINDOWS or DARWIN):
            self.iface = iface = conf.iface
        if promisc is None:
            promisc = conf.sniff_promisc
        if iface is not None:
            args.extend(['-i', network_name(iface)])
        if not promisc:
            args.append('-p')
        if not nofilter:
            if conf.except_filter:
                if filter:
                    filter = "(%s) and not (%s)" % (filter, conf.except_filter)
                else:
                    filter = "not (%s)" % conf.except_filter
        if filter is not None:
            args.append(filter)
        self.tcpdump_proc = tcpdump(
            None, prog=prog, args=args, getproc=True, quiet=quiet)
        self.reader = PcapReader(self.tcpdump_proc.stdout)
        self.ins = self.reader  # type: ignore

    def recv(self, x=MTU, **kwargs):
        # type: (int, **Any) -> Optional[Packet]
        return self.reader.recv(x, **kwargs)

    def close(self):
        # type: () -> None
        SuperSocket.close(self)
        self.tcpdump_proc.kill()

    @staticmethod
    def select(sockets, remain=None):
        # type: (List[SuperSocket], Optional[float]) -> List[SuperSocket]
        if (WINDOWS or DARWIN):
            return sockets
        return SuperSocket.select(sockets, remain=remain)


# More abstract objects

class IterSocket(SuperSocket):
    desc = "wrapper around an iterable"
    nonblocking_socket = True

    def __init__(self, obj):
        # type: (_PacketIterable) -> None
        if not obj:
            self.iter = iter([])  # type: Iterator[Packet]
        elif isinstance(obj, IterSocket):
            self.iter = obj.iter
        elif isinstance(obj, SndRcvList):
            def _iter(obj=cast(SndRcvList, obj)):
                # type: (SndRcvList) -> Iterator[Packet]
                for s, r in obj:
                    if s.sent_time:
                        s.time = s.sent_time
                    yield s
                    yield r
            self.iter = _iter()
        elif isinstance(obj, (list, PacketList)):
            if isinstance(obj[0], bytes):
                self.iter = iter(obj)
            else:
                self.iter = (y for x in obj for y in x)
        else:
            self.iter = obj.__iter__()

    @staticmethod
    def select(sockets, remain=None):
        # type: (List[SuperSocket], Any) -> List[SuperSocket]
        return sockets

    def recv(self, x=None, **kwargs):
        # type: (Optional[int], Any) -> Optional[Packet]
        try:
            pkt = next(self.iter)
            return pkt.__class__(bytes(pkt), **kwargs)
        except StopIteration:
            raise EOFError

    def close(self):
        # type: () -> None
        pass



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.status = skip

"""
Package of contrib automotive gm specific modules
that have to be loaded explicitly.
"""



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Fabian Wiche <f.wiche@gmx.de>
# Copyright (C) Tabea Spahn <tabea.spahn@e-mundo.de>

import argparse
import signal
import sys

from scapy.contrib.automotive.xcp.scanner import XCPOnCANScanner
from scapy.contrib.automotive.xcp.xcp import XCPOnCAN
from scapy.contrib.cansocket import CANSocket


class ScannerParams:
    def __init__(self):
        self.id_range = None
        self.sniff_time = None
        self.verbose = False
        self.channel = None
        self.broadcast = False


def signal_handler(sig, _frame):
    sys.stderr.write("Interrupting scan!\n")
    # Use same convention as the bash shell
    # 128+n where n is the fatal error signal
    # https://tldp.org/LDP/abs/html/exitcodes.html#EXITCODESREF
    sys.exit(128 + sig)


def init_socket(scan_params):
    print("Initializing socket for " + scan_params.channel)
    try:
        sock = CANSocket(scan_params.channel)
    except Exception as e:
        sys.stderr.write("\nSocket could not be created: " + str(e) + "\n")
        sys.exit(1)
    sock.basecls = XCPOnCAN
    return sock


def parse_inputs():
    scanner_params = ScannerParams()

    parser = argparse.ArgumentParser()
    parser.description = "Finds XCP slaves using the XCP Broadcast-CAN " \
                         "identifier."
    parser.add_argument('--start', '-s',
                        help='Start ID CAN (in hex).\n'
                             'If actual ID is unknown the scan will '
                             'test broadcast ids between --start and --end '
                             '(inclusive). Default: 0x00')
    parser.add_argument('--end', '-e',
                        help='End ID CAN (in hex).\n'
                             'If actual ID is unknown the scan will test '
                             'broadcast ids between --start and --end '
                             '(inclusive). Default: 0x7ff')
    parser.add_argument('--sniff_time', '-t',
                        help='Duration in milliseconds a sniff is waiting '
                             'for a response.', type=int, default=100)
    parser.add_argument('channel',
                        help='Linux SocketCAN interface name, e.g.: vcan0')
    parser.add_argument('--verbose', '-v', action="store_true",
                        help='Display information during scan')
    parser.add_argument('--broadcast', '-b', action="store_true",
                        help='Use Broadcast-message GetSlaveId instead of '
                             'default "Connect"')

    args = parser.parse_args()
    scanner_params.channel = args.channel
    scanner_params.verbose = args.verbose
    scanner_params.use_broadcast = args.broadcast
    scanner_params.sniff_time = float(args.sniff_time) / 1000

    start_id = int(args.start, 16) if args.start is not None else 0
    end_id = int(args.end, 16) if args.end is not None else 0x7ff

    if start_id > end_id:
        parser.error(
            "End identifier must not be smaller than the start identifier.")
        sys.exit(1)
    scanner_params.id_range = range(start_id, end_id + 1)

    return scanner_params


def main():
    scanner_params = parse_inputs()
    can_socket = init_socket(scanner_params)

    try:
        scanner = XCPOnCANScanner(can_socket,
                                  id_range=scanner_params.id_range,
                                  sniff_time=scanner_params.sniff_time,
                                  verbose=scanner_params.verbose)

        signal.signal(signal.SIGINT, signal_handler)

        results = scanner.scan_with_get_slave_id() \
            if scanner_params.broadcast \
            else scanner.scan_with_connect()  # Blocking

        if isinstance(results, list) and len(results) > 0:
            for r in results:
                print(r)
        else:
            print("Detected no XCP slave.")
    except Exception as err:
        sys.stderr.write(str(err) + "\n")
        sys.exit(1)
    finally:
        can_socket.close()


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Functions common to different architectures
"""

import ctypes
import re
import socket

from scapy.config import conf
from scapy.data import MTU, ARPHRD_TO_DLT, DLT_RAW_ALT, DLT_RAW
from scapy.error import Scapy_Exception, warning
from scapy.interfaces import network_name, resolve_iface, NetworkInterface
from scapy.libs.structures import bpf_program
from scapy.pton_ntop import inet_pton
from scapy.utils import decode_locale_str

# Type imports
import scapy
from typing import (
    List,
    Optional,
    Union,
)

# From if.h
_iff_flags = [
    "UP",
    "BROADCAST",
    "DEBUG",
    "LOOPBACK",
    "POINTTOPOINT",
    "NOTRAILERS",
    "RUNNING",
    "NOARP",
    "PROMISC",
    "ALLMULTI",
    "MASTER",
    "SLAVE",
    "MULTICAST",
    "PORTSEL",
    "AUTOMEDIA",
    "DYNAMIC",
    "LOWER_UP",
    "DORMANT",
    "ECHO"
]


def get_if_raw_addr(iff):
    # type: (Union[NetworkInterface, str]) -> bytes
    """Return the raw IPv4 address of interface"""
    iff = resolve_iface(iff)
    if not iff.ip:
        return b"\x00" * 4
    return inet_pton(socket.AF_INET, iff.ip)


# BPF HANDLERS


def compile_filter(filter_exp,  # type: str
                   iface=None,  # type: Optional[Union[str, 'scapy.interfaces.NetworkInterface']]  # noqa: E501
                   linktype=None,  # type: Optional[int]
                   promisc=False  # type: bool
                   ):
    # type: (...) -> bpf_program
    """Asks libpcap to parse the filter, then build the matching
    BPF bytecode.

    :param iface: if provided, use the interface to compile
    :param linktype: if provided, use the linktype to compile
    """
    try:
        from scapy.libs.winpcapy import (
            PCAP_ERRBUF_SIZE,
            pcap_open_live,
            pcap_compile,
            pcap_compile_nopcap,
            pcap_close
        )
    except OSError:
        raise ImportError(
            "libpcap is not available. Cannot compile filter !"
        )
    from ctypes import create_string_buffer
    bpf = bpf_program()
    bpf_filter = create_string_buffer(filter_exp.encode("utf8"))
    if not linktype:
        # Try to guess linktype to avoid root
        if not iface:
            if not conf.iface:
                raise Scapy_Exception(
                    "Please provide an interface or linktype!"
                )
            iface = conf.iface
        # Try to guess linktype to avoid requiring root
        try:
            arphd = resolve_iface(iface).type
            linktype = ARPHRD_TO_DLT.get(arphd)
        except Exception:
            # Failed to use linktype: use the interface
            pass
    if linktype is not None:
        # Some conversion aliases (e.g. linktype_to_dlt in libpcap)
        if linktype == DLT_RAW_ALT:
            linktype = DLT_RAW
        ret = pcap_compile_nopcap(
            MTU, linktype, ctypes.byref(bpf), bpf_filter, 1, -1
        )
    elif iface:
        err = create_string_buffer(PCAP_ERRBUF_SIZE)
        iface_b = create_string_buffer(network_name(iface).encode("utf8"))
        pcap = pcap_open_live(
            iface_b, MTU, promisc, 0, err
        )
        error = decode_locale_str(bytearray(err).strip(b"\x00"))
        if error:
            raise OSError(error)
        ret = pcap_compile(
            pcap, ctypes.byref(bpf), bpf_filter, 1, -1
        )
        pcap_close(pcap)
    if ret == -1:
        raise Scapy_Exception(
            "Failed to compile filter expression %s (%s)" % (filter_exp, ret)
        )
    return bpf


def free_filter(bp: bpf_program) -> None:
    """
    Free a bpf_program created with compile_filter
    """
    from scapy.libs.winpcapy import pcap_freecode
    pcap_freecode(ctypes.byref(bp))


#######
# DNS #
#######

def read_nameservers() -> List[str]:
    """Return the nameservers configured by the OS
    """
    try:
        with open('/etc/resolv.conf', 'r') as fd:
            return re.findall(r"nameserver\s+([^\s]+)", fd.read())
    except FileNotFoundError:
        warning("Could not retrieve the OS's nameserver !")
        return []



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>


"""A minimal implementation of the CANopen protocol, based on
Wireshark dissectors. See https://wiki.wireshark.org/CANopen

"""

import os
import gzip
import struct

from scapy.config import conf
from scapy.compat import chb, hex_bytes
from scapy.data import DLT_CAN_SOCKETCAN
from scapy.fields import FieldLenField, FlagsField, StrLenField, \
    ThreeBytesField, XBitField, ScalingField, ConditionalField, LenField, ShortField
from scapy.volatile import RandFloat, RandBinFloat
from scapy.packet import Packet, bind_layers
from scapy.layers.l2 import CookedLinux
from scapy.error import Scapy_Exception
from scapy.plist import PacketList
from scapy.supersocket import SuperSocket
from scapy.utils import _ByteStream

# Typing imports
from typing import (
    Tuple,
    Optional,
    Type,
    List,
    Union,
    Callable,
    IO,
    Any,
    cast,
)

__all__ = ["CAN", "SignalPacket", "SignalField", "LESignedSignalField",
           "LEUnsignedSignalField", "LEFloatSignalField", "BEFloatSignalField",
           "BESignedSignalField", "BEUnsignedSignalField", "rdcandump",
           "CandumpReader", "SignalHeader", "CAN_MTU", "CAN_MAX_IDENTIFIER",
           "CAN_MAX_DLEN", "CAN_INV_FILTER", "CANFD", "CAN_FD_MTU",
           "CAN_FD_MAX_DLEN"]

# CONSTANTS
CAN_MAX_IDENTIFIER = (1 << 29) - 1  # Maximum 29-bit identifier
CAN_MTU = 16
CAN_MAX_DLEN = 8
CAN_INV_FILTER = 0x20000000
CAN_FD_MTU = 72
CAN_FD_MAX_DLEN = 64

# Mimics the Wireshark CAN dissector parameter
# 'Byte-swap the CAN ID/flags field'.
# Set to True when working with PF_CAN sockets
conf.contribs['CAN'] = {'swap-bytes': False,
                        'remove-padding': True}


class CAN(Packet):
    """A implementation of CAN messages.

    Dissection of CAN messages from Wireshark captures and Linux PF_CAN sockets
    are supported from protocol specification.
    See https://wiki.wireshark.org/CANopen for further information on
    the Wireshark dissector. Linux PF_CAN and Wireshark use different
    endianness for the first 32 bit of a CAN message. This dissector can be
    configured for both use cases.

    Configuration ``swap-bytes``:
        Wireshark dissection:
            >>> conf.contribs['CAN']['swap-bytes'] = False

        PF_CAN Socket dissection:
            >>> conf.contribs['CAN']['swap-bytes'] = True

    Configuration ``remove-padding``:
    Linux PF_CAN Sockets always return 16 bytes per CAN frame receive.
    This implicates that CAN frames get padded from the Linux PF_CAN socket
    with zeros up to 8 bytes of data. The real length from the CAN frame on
    the wire is given by the length field. To obtain only the CAN frame from
    the wire, this additional padding has to be removed. Nevertheless, for
    corner cases, it might be useful to also get the padding. This can be
    configured through the **remove-padding** configuration.

    Truncate CAN frame based on length field:
        >>> conf.contribs['CAN']['remove-padding'] = True

    Show entire CAN frame received from socket:
        >>> conf.contribs['CAN']['remove-padding'] = False

    """
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        FieldLenField('length', None, length_of='data', fmt='B'),
        ThreeBytesField('reserved', 0),
        StrLenField('data', b'', length_from=lambda pkt: int(pkt.length)),
    ]

    @classmethod
    def dispatch_hook(cls,
                      _pkt=None,  # type: Optional[bytes]
                      *args,  # type: Any
                      **kargs  # type: Any
                      ):  # type: (...) -> Type[Packet]
        if _pkt:
            fdf_set = len(_pkt) > 5 and _pkt[5] & 0x04 and \
                not _pkt[5] & 0xf8
            if fdf_set:
                return CANFD
            elif len(_pkt) > 4 and _pkt[4] > 8:
                return CANFD
        return CAN

    @staticmethod
    def inv_endianness(pkt):
        # type: (bytes) -> bytes
        """Invert the order of the first four bytes of a CAN packet

        This method is meant to be used specifically to convert a CAN packet
        between the pcap format and the SocketCAN format

        :param pkt: bytes str of the CAN packet
        :return: bytes str with the first four bytes swapped
        """
        len_partial = len(pkt) - 4  # len of the packet, CAN ID excluded
        return struct.pack('<I{}s'.format(len_partial),
                           *struct.unpack('>I{}s'.format(len_partial), pkt))

    def pre_dissect(self, s):
        # type: (bytes) -> bytes
        """Implements the swap-bytes functionality when dissecting """
        if conf.contribs['CAN']['swap-bytes']:
            data = CAN.inv_endianness(s)  # type: bytes
            return data
        return s

    def post_dissect(self, s):
        # type: (bytes) -> bytes
        self.raw_packet_cache = None  # Reset packet to allow post_build
        return s

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes
        """Implements the swap-bytes functionality for Packet build.

        This is based on a copy of the Packet.self_build default method.
        The goal is to affect only the CAN layer data and keep
        under layers (e.g CookedLinux) unchanged
        """
        if conf.contribs['CAN']['swap-bytes']:
            data = CAN.inv_endianness(pkt)  # type: bytes
            return data + pay
        return pkt + pay

    def extract_padding(self, p):
        # type: (bytes) -> Tuple[bytes, Optional[bytes]]
        if conf.contribs['CAN']['remove-padding']:
            return b'', None
        else:
            return b'', p


conf.l2types.register(DLT_CAN_SOCKETCAN, CAN)
bind_layers(CookedLinux, CAN, proto=12)


class CANFD(CAN):
    """
    This class is used for distinction of CAN FD packets.
    """
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        FieldLenField('length', None, length_of='data', fmt='B'),
        FlagsField('fd_flags', 4, 8, ['bit_rate_switch',
                                      'error_state_indicator',
                                      'fd_frame']),
        ShortField('reserved', 0),
        StrLenField('data', b'', length_from=lambda pkt: int(pkt.length)),
    ]

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes

        data = super(CANFD, self).post_build(pkt, pay)

        length = data[4]

        if 8 < length <= 24:
            wire_length = length + (-length) % 4
        elif 24 < length <= 64:
            wire_length = length + (-length) % 8
        elif length > 64:
            raise NotImplementedError
        else:
            wire_length = length

        pad = b"\x00" * (wire_length - length)

        return data[0:4] + chb(wire_length) + data[5:] + pad


bind_layers(CookedLinux, CANFD, proto=13)


class SignalField(ScalingField):
    """SignalField is a base class for signal data, usually transmitted from
    CAN messages in automotive applications. Most vehicle manufacturers
    describe their vehicle internal signals by so called data base CAN (DBC)
    files. All necessary functions to easily create Scapy dissectors similar
    to signal descriptions from DBC files are provided by this base class.

    SignalField instances should only be used together with SignalPacket
    classes since SignalPackets enforce length checks for CAN messages.

    """
    __slots__ = ["start", "size"]

    def __init__(self, name, default, start, size, scaling=1, unit="",
                 offset=0, ndigits=3, fmt="B"):
        # type: (str, Union[int, float], int, int, Union[int, float], str, Union[int, float], int, str) -> None  # noqa: E501
        ScalingField.__init__(self, name, default, scaling, unit, offset,
                              ndigits, fmt)
        self.start = start
        self.size = abs(size)

        if fmt[-1] == "f" and self.size != 32:
            raise Scapy_Exception("SignalField size has to be 32 for floats")

    _lookup_table = [7, 6, 5, 4, 3, 2, 1, 0,
                     15, 14, 13, 12, 11, 10, 9, 8,
                     23, 22, 21, 20, 19, 18, 17, 16,
                     31, 30, 29, 28, 27, 26, 25, 24,
                     39, 38, 37, 36, 35, 34, 33, 32,
                     47, 46, 45, 44, 43, 42, 41, 40,
                     55, 54, 53, 52, 51, 50, 49, 48,
                     63, 62, 61, 60, 59, 58, 57, 56]

    @staticmethod
    def _msb_lookup(start):
        # type: (int) -> int
        try:
            return SignalField._lookup_table.index(start)
        except ValueError:
            raise Scapy_Exception("Only 64 bits for all SignalFields "
                                  "are supported")

    @staticmethod
    def _lsb_lookup(start, size):
        # type: (int, int) -> int
        try:
            return SignalField._lookup_table[SignalField._msb_lookup(start) +
                                             size - 1]
        except IndexError:
            raise Scapy_Exception("Only 64 bits for all SignalFields "
                                  "are supported")

    @staticmethod
    def _convert_to_unsigned(number, bit_length):
        # type: (int, int) -> int
        if number & (1 << (bit_length - 1)):
            mask = (2 ** bit_length)  # type: int
            return mask + number
        return number

    @staticmethod
    def _convert_to_signed(number, bit_length):
        # type: (int, int) -> int
        mask = (2 ** bit_length) - 1  # type: int
        if number & (1 << (bit_length - 1)):
            return number | ~mask
        return number & mask

    def _is_little_endian(self):
        # type: () -> bool
        return self.fmt[0] == "<"

    def _is_signed_number(self):
        # type: () -> bool
        return self.fmt[-1].islower()

    def _is_float_number(self):
        # type: () -> bool
        return self.fmt[-1] == "f"

    def addfield(self, pkt, s, val):
        # type: (Packet, bytes, Optional[Union[int, float]]) -> bytes
        if not isinstance(pkt, SignalPacket):
            raise Scapy_Exception("Only use SignalFields in a SignalPacket")

        val = self.i2m(pkt, val)

        if self._is_little_endian():
            msb_pos = self.start + self.size - 1
            lsb_pos = self.start
            shift = lsb_pos
            fmt = "<Q"
        else:
            msb_pos = self.start
            lsb_pos = self._lsb_lookup(self.start, self.size)
            shift = (64 - self._msb_lookup(msb_pos) - self.size)
            fmt = ">Q"

        field_len = max(msb_pos, lsb_pos) // 8 + 1
        if len(s) < field_len:
            s += b"\x00" * (field_len - len(s))

        if self._is_float_number():
            int_val = struct.unpack(self.fmt[0] + "I",
                                    struct.pack(self.fmt, val))[0]  # type: int
        elif self._is_signed_number():
            int_val = self._convert_to_unsigned(int(val), self.size)
        else:
            int_val = cast(int, val)

        pkt_val = struct.unpack(fmt, (s + b"\x00" * 8)[:8])[0]
        pkt_val |= int_val << shift
        tmp_s = struct.pack(fmt, pkt_val)
        return tmp_s[:len(s)]

    def getfield(self, pkt, s):
        # type: (Packet, bytes) -> Tuple[bytes, Union[int, float]]
        if not isinstance(pkt, SignalPacket):
            raise Scapy_Exception("Only use SignalFields in a SignalPacket")

        if isinstance(s, tuple):
            s, _ = s

        if self._is_little_endian():
            msb_pos = self.start + self.size - 1
            lsb_pos = self.start
            shift = self.start
            fmt = "<Q"
        else:
            msb_pos = self.start
            lsb_pos = self._lsb_lookup(self.start, self.size)
            shift = (64 - self._msb_lookup(self.start) - self.size)
            fmt = ">Q"

        field_len = max(msb_pos, lsb_pos) // 8 + 1

        if pkt.wirelen is None:
            pkt.wirelen = field_len

        pkt.wirelen = max(pkt.wirelen, field_len)

        fld_val = struct.unpack(fmt, (s + b"\x00" * 8)[:8])[0] >> shift
        fld_val &= ((1 << self.size) - 1)

        if self._is_float_number():
            fld_val = struct.unpack(self.fmt,
                                    struct.pack(self.fmt[0] + "I", fld_val))[0]
        elif self._is_signed_number():
            fld_val = self._convert_to_signed(fld_val, self.size)

        return s, self.m2i(pkt, fld_val)

    def randval(self):
        # type: () -> Union[RandBinFloat, RandFloat]
        if self._is_float_number():
            return RandBinFloat(0, 0)

        if self._is_signed_number():
            min_val = -2**(self.size - 1)
            max_val = 2**(self.size - 1) - 1
        else:
            min_val = 0
            max_val = 2 ** self.size - 1

        min_val = round(min_val * self.scaling + self.offset, self.ndigits)
        max_val = round(max_val * self.scaling + self.offset, self.ndigits)

        return RandFloat(min(min_val, max_val), max(min_val, max_val))

    def i2len(self, pkt, x):
        # type: (Packet, Any) -> int
        return int(float(self.size) / 8)


class LEUnsignedSignalField(SignalField):
    def __init__(self, name, default, start, size, scaling=1, unit="",
                 offset=0, ndigits=3):
        # type: (str, Union[int, float], int, int, Union[int, float], str, Union[int, float], int) -> None  # noqa: E501
        SignalField.__init__(self, name, default, start, size,
                             scaling, unit, offset, ndigits, "<B")


class LESignedSignalField(SignalField):
    def __init__(self, name, default, start, size, scaling=1, unit="",
                 offset=0, ndigits=3):
        # type: (str, Union[int, float], int, int, Union[int, float], str, Union[int, float], int) -> None  # noqa: E501
        SignalField.__init__(self, name, default, start, size,
                             scaling, unit, offset, ndigits, "<b")


class BEUnsignedSignalField(SignalField):
    def __init__(self, name, default, start, size, scaling=1, unit="",
                 offset=0, ndigits=3):
        # type: (str, Union[int, float], int, int, Union[int, float], str, Union[int, float], int) -> None  # noqa: E501
        SignalField.__init__(self, name, default, start, size,
                             scaling, unit, offset, ndigits, ">B")


class BESignedSignalField(SignalField):
    def __init__(self, name, default, start, size, scaling=1, unit="",
                 offset=0, ndigits=3):
        # type: (str, Union[int, float], int, int, Union[int, float], str, Union[int, float], int) -> None  # noqa: E501
        SignalField.__init__(self, name, default, start, size,
                             scaling, unit, offset, ndigits, ">b")


class LEFloatSignalField(SignalField):
    def __init__(self, name, default, start, scaling=1, unit="",
                 offset=0, ndigits=3):
        # type: (str, Union[int, float], int, Union[int, float], str, Union[int, float], int) -> None  # noqa: E501
        SignalField.__init__(self, name, default, start, 32,
                             scaling, unit, offset, ndigits, "<f")


class BEFloatSignalField(SignalField):
    def __init__(self, name, default, start, scaling=1, unit="",
                 offset=0, ndigits=3):
        # type: (str, Union[int, float], int, Union[int, float], str, Union[int, float], int) -> None  # noqa: E501
        SignalField.__init__(self, name, default, start, 32,
                             scaling, unit, offset, ndigits, ">f")


class SignalPacket(Packet):
    """Special implementation of Packet.

    This class enforces the correct wirelen of a CAN message for
    signal transmitting in automotive applications.
    Furthermore, the dissection order of SignalFields in fields_desc is
    deduced by the start index of a field.
    """

    def pre_dissect(self, s):
        # type: (bytes) -> bytes
        if not all(isinstance(f, SignalField) or
                   (isinstance(f, ConditionalField) and
                    isinstance(f.fld, SignalField))
                   for f in self.fields_desc):
            raise Scapy_Exception("Use only SignalFields in a SignalPacket")
        return s

    def post_dissect(self, s):
        # type: (bytes) -> bytes
        """SignalFields can be dissected on packets with unordered fields.

        The order of SignalFields is defined from the start parameter.
        After a build, the consumed bytes of the length of all SignalFields
        have to be removed from the SignalPacket.
        """
        if self.wirelen is not None and self.wirelen > 8:
            raise Scapy_Exception("Only 64 bits for all SignalFields "
                                  "are supported")
        self.raw_packet_cache = None  # Reset packet to allow post_build
        return s[self.wirelen:]


class SignalHeader(CAN):
    """Special implementation of a CAN Packet to allow dynamic binding.

    This class can be provided to CANSockets as basecls.

    Example:
        >>> class floatSignals(SignalPacket):
        >>>     fields_desc = [
        >>>         LEFloatSignalField("floatSignal2", default=0, start=32),
        >>>         BEFloatSignalField("floatSignal1", default=0, start=7)]
        >>>
        >>> bind_layers(SignalHeader, floatSignals, identifier=0x321)
        >>>
        >>> dbc_sock = CANSocket("can0", basecls=SignalHeader)

    All CAN messages received from this dbc_sock CANSocket will be interpreted
    as SignalHeader. Through Scapys ``bind_layers`` mechanism, all CAN messages
    with CAN identifier 0x321 will interpret the payload bytes of these
    CAN messages as floatSignals packet.
    """
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        LenField('length', None, fmt='B'),
        FlagsField('fd_flags', 0, 8, ['bit_rate_switch',
                                      'error_state_indicator',
                                      'fd_frame']),
        ShortField('reserved', 0)
    ]

    @classmethod
    def dispatch_hook(cls,
                      _pkt=None,  # type: Optional[bytes]
                      *args,  # type: Any
                      **kargs  # type: Any
                      ):  # type: (...) -> Type[Packet]
        return SignalHeader

    def extract_padding(self, s):
        # type: (bytes) -> Tuple[bytes, Optional[bytes]]
        return s, None


def rdcandump(filename, count=-1, interface=None):
    # type: (str, int, Optional[str]) -> PacketList
    """ Read a candump log file and return a packet list.

    :param filename: Filename of the file to read from.
                     Also gzip files are accepted.
    :param count: Read only <count> packets. Specify -1 to read all packets.
    :param interface: Return only packets from a specified interface
    :return: A PacketList object containing the read files
    """
    with CandumpReader(filename, interface) as fdesc:
        return fdesc.read_all(count=count)


class CandumpReader:
    """A stateful candump reader. Each packet is returned as a CAN packet.

    Creates a CandumpReader object

    :param filename: filename of a candump logfile, compressed or
                     uncompressed, or a already opened file object.
    :param interface: Name of a interface, if candump contains messages
                      of multiple interfaces and only one messages from a
                      specific interface are wanted.
    """

    nonblocking_socket = True

    def __init__(self, filename, interface=None):
        # type: (str, Optional[Union[List[str], str]]) -> None
        self.filename, self.f = self.open(filename)
        self.ifilter = None  # type: Optional[List[str]]
        if interface is not None:
            if isinstance(interface, str):
                self.ifilter = [interface]
            else:
                self.ifilter = interface

    def __iter__(self):
        # type: () -> CandumpReader
        return self

    @staticmethod
    def open(filename):
        # type: (Union[IO[bytes], str]) -> Tuple[str, _ByteStream]
        """Open function to handle three types of input data.

        If filename of a regular candump log file is provided, this function
        opens the file and returns the file object.
        If filename of a gzip compressed candump log file is provided, the
        required gzip open function is used to obtain the necessary file
        object, which gets returned.
        If a fileobject or ByteIO is provided, the filename is gathered for
        internal use. No further steps are performed on this object.

        :param filename: Can be a string, specifying a candump log file or a
                         gzip compressed candump log file. Also already opened
                         file objects are allowed.
        :return: A opened file object for further use.
        """
        """Open (if necessary) filename."""
        if isinstance(filename, str):
            try:
                fdesc = gzip.open(filename, "rb")  # type: _ByteStream
                # try read to cause exception
                fdesc.read(1)
                fdesc.seek(0)
            except IOError:
                fdesc = open(filename, "rb")
            return filename, fdesc
        else:
            name = getattr(filename, "name", "No name")
            return name, filename

    def next(self):
        # type: () -> Packet
        """Implements the iterator protocol on a set of packets

        :return: Next readable CAN Packet from the specified file
        """
        try:
            pkt = None
            while pkt is None:
                pkt = self.read_packet()
        except EOFError:
            raise StopIteration

        return pkt
    __next__ = next

    def read_packet(self, size=CAN_MTU):
        # type: (int) -> Optional[Packet]
        """Read a packet from the specified file.

        This function will raise EOFError when no more packets are available.

        :param size: Not used. Just here to follow the function signature for
                     SuperSocket emulation.
        :return: A single packet read from the file or None if filters apply
        """
        line = self.f.readline()
        line = line.lstrip()
        if len(line) < 16:
            raise EOFError

        is_log_file_format = line[0] == ord(b"(")
        fd_flags = None
        if is_log_file_format:
            t_b, intf, f = line.split()
            if b'##' in f:
                idn, data = f.split(b'##')
                fd_flags = data[0]
                data = data[1:]
            else:
                idn, data = f.split(b'#')
            le = None
            t = float(t_b[1:-1])  # type: Optional[float]
        else:
            h, data = line.split(b']')
            intf, idn, le = h.split()
            t = None

        if self.ifilter is not None and \
                intf.decode('ASCII') not in self.ifilter:
            return None

        data = data.replace(b' ', b'')
        data = data.strip()

        if len(data) <= 8 and fd_flags is None:
            pkt = CAN(identifier=int(idn, 16), data=hex_bytes(data))
        else:
            pkt = CANFD(identifier=int(idn, 16), fd_flags=fd_flags,
                        data=hex_bytes(data))

        if le is not None:
            pkt.length = int(le[1:])
        else:
            pkt.length = len(pkt.data)

        if len(idn) > 3:
            pkt.flags = 0b100

        if t is not None:
            pkt.time = t

        return pkt

    def dispatch(self, callback):
        # type: (Callable[[Packet], None]) -> None
        """Call the specified callback routine for each packet read

        This is just a convenience function for the main loop
        that allows for easy launching of packet processing in a
        thread.
        """
        for p in self:
            callback(p)

    def read_all(self, count=-1):
        # type: (int) -> PacketList
        """Read a specific number or all packets from a candump file.

        :param count: Specify a specific number of packets to be read.
                      All packets can be read by count=-1.
        :return: A PacketList object containing read CAN messages
        """
        res = []
        while count != 0:
            try:
                p = self.read_packet()
                if p is None:
                    continue
            except EOFError:
                break
            count -= 1
            res.append(p)
        return PacketList(res, name=os.path.basename(self.filename))

    def recv(self, size=CAN_MTU):
        # type: (int) -> Optional[Packet]
        """Emulation of SuperSocket"""
        try:
            return self.read_packet(size=size)
        except EOFError:
            return None

    def fileno(self):
        # type: () -> int
        """Emulation of SuperSocket"""
        return self.f.fileno()

    @property
    def closed(self):
        # type: () -> bool
        return self.f.closed

    def close(self):
        # type: () -> Any
        """Emulation of SuperSocket"""
        return self.f.close()

    def __enter__(self):
        # type: () -> CandumpReader
        return self

    def __exit__(self, exc_type, exc_value, tracback):
        # type: (Optional[Type[BaseException]], Optional[BaseException], Optional[Any]) -> None  # noqa: E501
        self.close()

    @staticmethod
    def select(sockets, remain=None):
        # type: (List[SuperSocket], Optional[int]) -> List[SuperSocket]
        """Emulation of SuperSocket"""
        return [s for s in sockets if isinstance(s, CandumpReader) and
                not s.closed]



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy RPC
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

# ms-drsr.idl compiled on 06/07/2025
# This file is a stripped version ! Use scapy-rpc for the full.
"""
RPC definitions for the following interfaces:
- drsuapi (v4.0): e3514235-4b06-11d1-ab04-00c04fc2dcd2
This file is auto-generated by midl-to-scapy, do not modify.
"""

import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRConfFieldListField,
    NDRConfPacketListField,
    NDRConfStrLenField,
    NDRConfVarStrNullField,
    NDRConfVarStrNullFieldUtf16,
    NDRContextHandle,
    NDRFullEmbPointerField,
    NDRFullPointerField,
    NDRIntField,
    NDRPacketField,
    NDRShortField,
    NDRUnionField,
    register_dcerpc_interface,
)


class UUID(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("Data1", 0),
        NDRShortField("Data2", 0),
        NDRShortField("Data3", 0),
        StrFixedLenField("Data4", "", length=8),
    ]


class DRS_EXTENSIONS(NDRPacket):
    ALIGNMENT = (4, 8)
    DEPORTED_CONFORMANTS = ["rgb"]
    fields_desc = [
        NDRIntField("cb", None, size_of="rgb"),
        NDRConfStrLenField(
            "rgb", "", size_is=lambda pkt: pkt.cb, conformant_in_struct=True
        ),
    ]


class IDL_DRSBind_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRPacketField("puuidClientDsa", UUID(), UUID)),
        NDRFullPointerField(
            NDRPacketField("pextClient", DRS_EXTENSIONS(), DRS_EXTENSIONS)
        ),
    ]


class IDL_DRSBind_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(
            NDRPacketField("ppextServer", DRS_EXTENSIONS(), DRS_EXTENSIONS)
        ),
        NDRPacketField("phDrs", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class IDL_DRSUnbind_Request(NDRPacket):
    fields_desc = [NDRPacketField("phDrs", NDRContextHandle(), NDRContextHandle)]


class IDL_DRSUnbind_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("phDrs", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class DRS_MSG_CRACKREQ_V1(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("CodePage", 0),
        NDRIntField("LocaleId", 0),
        NDRIntField("dwFlags", 0),
        NDRIntField("formatOffered", 0),
        NDRIntField("formatDesired", 0),
        NDRIntField("cNames", None, size_of="rpNames"),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "rpNames",
                [],
                NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("rpNames", "")),
                size_is=lambda pkt: pkt.cNames,
            )
        ),
    ]


class PDS_NAME_RESULT_ITEMW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("status", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("pDomain", "")),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("pName", "")),
    ]


class DS_NAME_RESULTW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("cItems", None, size_of="rItems"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "rItems",
                [PDS_NAME_RESULT_ITEMW()],
                PDS_NAME_RESULT_ITEMW,
                size_is=lambda pkt: pkt.cItems,
            )
        ),
    ]


class DRS_MSG_CRACKREPLY_V1(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(
            NDRPacketField("pResult", DS_NAME_RESULTW(), DS_NAME_RESULTW)
        )
    ]


class IDL_DRSCrackNames_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hDrs", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwInVersion", 0),
        NDRUnionField(
            [
                (
                    NDRPacketField(
                        "pmsgIn", DRS_MSG_CRACKREQ_V1(), DRS_MSG_CRACKREQ_V1
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInVersion", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("pmsgIn", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class IDL_DRSCrackNames_Response(NDRPacket):
    fields_desc = [
        NDRIntField("pdwOutVersion", 0),
        NDRUnionField(
            [
                (
                    NDRPacketField(
                        "pmsgOut", DRS_MSG_CRACKREPLY_V1(), DRS_MSG_CRACKREPLY_V1
                    ),
                    (
                        (lambda pkt: getattr(pkt, "pdwOutVersion", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("pmsgOut", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


DRSUAPI_OPNUMS = {
    0: DceRpcOp(IDL_DRSBind_Request, IDL_DRSBind_Response),
    1: DceRpcOp(IDL_DRSUnbind_Request, IDL_DRSUnbind_Response),
    12: DceRpcOp(IDL_DRSCrackNames_Request, IDL_DRSCrackNames_Response),
}
register_dcerpc_interface(
    name="drsuapi",
    uuid=uuid.UUID("e3514235-4b06-11d1-ab04-00c04fc2dcd2"),
    version="4.0",
    opnums=DRSUAPI_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Andreas Korb <andreas.d.korb@gmail.com>
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.status = skip

from scapy.fields import StrFixedLenField, FlagsField, ScalingField, BitField
from scapy.contrib.automotive.obd.packet import OBD_Packet


# See https://en.wikipedia.org/wiki/OBD-II_PIDs for further information
# PID = Parameter IDentification

class OBD_PID80(OBD_Packet):
    name = "PID_80_PIDsSupported"
    fields_desc = [
        FlagsField('supported_pids', 0, 32, [
            'PIDA0',
            'PID9F',
            'PID9E',
            'PID9D',
            'PID9C',
            'PID9B',
            'PID9A',
            'PID99',
            'PID98',
            'PID97',
            'PID96',
            'PID95',
            'PID94',
            'PID93',
            'PID92',
            'PID91',
            'PID90',
            'PID8F',
            'PID8E',
            'PID8D',
            'PID8C',
            'PID8B',
            'PID8A',
            'PID89',
            'PID88',
            'PID87',
            'PID86',
            'PID85',
            'PID84',
            'PID83',
            'PID82',
            'PID81'
        ])
    ]


class OBD_PID81(OBD_Packet):
    name = "PID_81_EngineRunTimeForAuxiliaryEmissionsControlDevice"
    fields_desc = [
        BitField('reserved', 0, 3),
        BitField('total_run_time_with_ei_aecd5_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd4_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd3_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd2_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd1_supported', 0, 1),
        ScalingField('total_run_time_with_ei_aecd1', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd2', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd3', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd4', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd5', 0, unit='sec',
                     fmt='Q'),
    ]


class OBD_PID82(OBD_Packet):
    name = "PID_82_EngineRunTimeForAuxiliaryEmissionsControlDevice"
    fields_desc = [
        BitField('reserved', 0, 3),
        BitField('total_run_time_with_ei_aecd10_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd9_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd8_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd7_supported', 0, 1),
        BitField('total_run_time_with_ei_aecd6_supported', 0, 1),
        ScalingField('total_run_time_with_ei_aecd6', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd7', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd8', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd9', 0, unit='sec',
                     fmt='Q'),
        ScalingField('total_run_time_with_ei_aecd10', 0, unit='sec',
                     fmt='Q'),
    ]


class OBD_PID83(OBD_Packet):
    name = "PID_83_NOxSensor"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('nox_sensor_concentration_bank2_sensor1_supported', 0, 1),
        BitField('nox_sensor_concentration_bank1_sensor1_supported', 0, 1),
        ScalingField('nox_sensor_concentration_bank1_sensor1', 0, unit='ppm',
                     fmt='H'),
        ScalingField('nox_sensor_concentration_bank2_sensor1', 0, unit='ppm',
                     fmt='H'),
    ]


class OBD_PID84(OBD_Packet):
    name = "PID_84_ManifoldSurfaceTemperature"
    fields_desc = [
        StrFixedLenField('data', b'', 1)
    ]


class OBD_PID85(OBD_Packet):
    name = "PID_85_NoxReagentSystem"
    fields_desc = [
        StrFixedLenField('data', b'', 10)
    ]


class OBD_PID86(OBD_Packet):
    name = "PID_86_ParticulateMatterSensor"
    fields_desc = [
        StrFixedLenField('data', b'', 5)
    ]


class OBD_PID87(OBD_Packet):
    name = "PID_87_IntakeManifoldAbsolutePressure"
    fields_desc = [
        StrFixedLenField('data', b'', 5)
    ]


class OBD_PID88(OBD_Packet):
    name = "PID_88_ScrInduceSystem"
    fields_desc = [
        StrFixedLenField('data', b'', 13)
    ]


class OBD_PID89(OBD_Packet):
    # 11 - 15
    name = "PID_89_RunTimeForAecd"
    fields_desc = [
        StrFixedLenField('data', b'', 41)
    ]


class OBD_PID8A(OBD_Packet):
    # 16 - 20
    name = "PID_8A_RunTimeForAecd"
    fields_desc = [
        StrFixedLenField('data', b'', 41)
    ]


class OBD_PID8B(OBD_Packet):
    name = "PID_8B_DieselAftertreatment"
    fields_desc = [
        StrFixedLenField('data', b'', 7)
    ]


class OBD_PID8C(OBD_Packet):
    name = "PID_8C_O2Sensor"
    fields_desc = [
        StrFixedLenField('data', b'', 16)
    ]


class OBD_PID8D(OBD_Packet):
    name = "PID_8D_ThrottlePositionG"
    fields_desc = [
        StrFixedLenField('data', b'', 1)
    ]


class OBD_PID8E(OBD_Packet):
    name = "PID_8E_EngineFrictionPercentTorque"
    fields_desc = [
        StrFixedLenField('data', b'', 1)
    ]


class OBD_PID8F(OBD_Packet):
    name = "PID_8F_PmSensorBank1And2"
    fields_desc = [
        StrFixedLenField('data', b'', 5)
    ]


class OBD_PID90(OBD_Packet):
    name = "PID_90_WwhObdVehicleObdSystemInformation"
    fields_desc = [
        StrFixedLenField('data', b'', 3)
    ]


class OBD_PID91(OBD_Packet):
    name = "PID_91_WwhObdVehicleObdSystemInformation"
    fields_desc = [
        StrFixedLenField('data', b'', 5)
    ]


class OBD_PID92(OBD_Packet):
    name = "PID_92_FuelSystemControl"
    fields_desc = [
        StrFixedLenField('data', b'', 2)
    ]


class OBD_PID93(OBD_Packet):
    name = "PID_93_WwhObdVehicleObdCountersSupport"
    fields_desc = [
        StrFixedLenField('data', b'', 3)
    ]


class OBD_PID94(OBD_Packet):
    name = "PID_94_NoxWarningAndInducementSystem"
    fields_desc = [
        StrFixedLenField('data', b'', 12)
    ]


class OBD_PID98(OBD_Packet):
    name = "PID_98_ExhaustGasTemperatureSensor"
    fields_desc = [
        StrFixedLenField('data', b'', 9)
    ]


class OBD_PID99(OBD_Packet):
    name = "PID_99_ExhaustGasTemperatureSensor"
    fields_desc = [
        StrFixedLenField('data', b'', 9)
    ]


class OBD_PID9A(OBD_Packet):
    name = "PID_9A_HybridEvVehicleSystemDataBatteryVoltage"
    fields_desc = [
        StrFixedLenField('data', b'', 6)
    ]


class OBD_PID9B(OBD_Packet):
    name = "PID_9B_DieselExhaustFluidSensorData"
    fields_desc = [
        StrFixedLenField('data', b'', 4)
    ]


class OBD_PID9C(OBD_Packet):
    name = "PID_9C_O2SensorData"
    fields_desc = [
        StrFixedLenField('data', b'', 17)
    ]


class OBD_PID9D(OBD_Packet):
    name = "PID_9D_EngineFuelRate"
    fields_desc = [
        StrFixedLenField('data', b'', 4)
    ]


class OBD_PID9E(OBD_Packet):
    name = "PID_9E_EngineExhaustFlowRate"
    fields_desc = [
        StrFixedLenField('data', b'', 2)
    ]


class OBD_PID9F(OBD_Packet):
    name = "PID_9F_FuelSystemPercentageUse"
    fields_desc = [
        StrFixedLenField('data', b'', 9)
    ]



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) 2021 Trend Micro Incorporated
# Copyright (C) 2021 Alias Robotics S.L.

"""
Real-Time Publish-Subscribe Protocol (RTPS) dissection
"""

# scapy.contrib.description = RTPS PID type definitions
# scapy.contrib.status = library

import random
import struct

from scapy.fields import (
    IntField,
    PacketField,
    PacketListField,
    ShortField,
    StrLenField,
    XIntField,
    XShortField,
    XStrFixedLenField,
)
from scapy.packet import Packet

from scapy.contrib.rtps.common_types import (
    STR_MAX_LEN,
    EField,
    EPacket,
    GUIDPacket,
    LeaseDurationPacket,
    LocatorPacket,
    ProductVersionPacket,
    ProtocolVersionPacket,
    TransportInfoPacket,
    VendorIdPacket,
    FORMAT_LE,
)


class ParameterIdField(XShortField):
    _valid_ids = [
        0x0002,
        0x0004,
        0x0005,
        0x0006,
        0x0007,
        0x000B,
        0x000C,
        0x000D,
        0x000E,
        0x000F,
        0x0011,
        0x0015,
        0x0016,
        0x001A,
        0x001B,
        0x001D,
        0x001E,
        0x001F,
        0x0021,
        0x0023,
        0x0025,
        0x0027,
        0x0029,
        0x002B,
        0x002C,
        0x002D,
        0x002E,
        0x002F,
        0x0030,
        0x0031,
        0x0032,
        0x0033,
        0x0034,
        0x0035,
        0x0040,
        0x0041,
        0x0043,
        0x0044,
        0x0045,
        0x0046,
        0x0048,
        0x0049,
        0x0050,
        0x0052,
        0x0053,
        0x0058,
        0x0059,
        0x005A,
        0x0060,
        0x0062,
        0x0070,
        0x0071,
        0x0077,
        0x4014,
        0x8000,
        0x8001,
        0x800f,
        0x8010,
        0x8016,
        0x8017
    ]

    def randval(self):
        return random.choice(self._valid_ids)


class PIDPacketBase(Packet):
    name = "PID Base Packet"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        StrLenField(
            "parameterData",
            "",
            length_from=lambda x: x.parameterLength,
            max_length=STR_MAX_LEN,
        ),
    ]

    def extract_padding(self, p):
        return b"", p


class PID_PAD(PIDPacketBase):
    name = "PID_PAD"
    fields_desc = [
        StrLenField(
            "parameterId", "",
            length_from=lambda x: 2,
            max_length=STR_MAX_LEN)
    ]


class PID_SENTINEL(PIDPacketBase):
    name = "PID_SENTINEL"


class PID_USER_DATA(PIDPacketBase):
    name = "PID_USER_DATA"


class PID_TOPIC_NAME(PIDPacketBase):
    name = "PID_TOPIC_NAME"


class PID_TYPE_NAME(PIDPacketBase):
    name = "PID_TYPE_NAME"


class PID_GROUP_DATA(PIDPacketBase):
    name = "PID_GROUP_DATA"


class PID_TOPIC_DATA(PIDPacketBase):
    name = "PID_TOPIC_DATA"


class PID_DURABILITY(PIDPacketBase):
    name = "PID_DURABILITY"


class PID_DURABILITY_SERVICE(PIDPacketBase):
    name = "PID_DURABILITY_SERVICE"


class PID_DEADLINE(PIDPacketBase):
    name = "PID_DEADLINE"


class PID_LATENCY_BUDGET(PIDPacketBase):
    name = "PID_LATENCY_BUDGET"


class PID_LIVELINESS(PIDPacketBase):
    name = "PID_LIVELINESS"


class PID_RELIABILITY(PIDPacketBase):
    name = "PID_RELIABILITY"


class PID_LIFESPAN(PIDPacketBase):
    name = "PID_LIFESPAN"


class PID_DESTINATION_ORDER(PIDPacketBase):
    name = "PID_DESTINATION_ORDER"


class PID_HISTORY(PIDPacketBase):
    name = "PID_HISTORY"


class PID_RESOURCE_LIMITS(PIDPacketBase):
    name = "PID_RESOURCE_LIMITS"


class PID_OWNERSHIP(PIDPacketBase):
    name = "PID_OWNERSHIP"


class PID_OWNERSHIP_STRENGTH(PIDPacketBase):
    name = "PID_OWNERSHIP_STRENGTH"


class PID_PRESENTATION(PIDPacketBase):
    name = "PID_PRESENTATION"


class PID_PARTITION(PIDPacketBase):
    name = "PID_PARTITION"


class PID_TIME_BASED_FILTER(PIDPacketBase):
    name = "PID_TIME_BASED_FILTER"


class PID_TRANSPORT_PRIO(PIDPacketBase):
    name = "PID_TRANSPORT_PRIO"


class PID_PROTOCOL_VERSION(PIDPacketBase):
    name = "PID_PROTOCOL_VERSION"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("protocolVersion", "", ProtocolVersionPacket),
        StrLenField(
            "padding",
            "",
            length_from=lambda x: x.parameterLength - 2,
            max_length=STR_MAX_LEN,
        ),
    ]


class PID_VENDOR_ID(PIDPacketBase):
    name = "PID_VENDOR_ID"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("vendorId", "", VendorIdPacket),
        StrLenField(
            "padding",
            "",
            length_from=lambda x: x.parameterLength - 2,
            max_length=STR_MAX_LEN,
        ),
    ]


class PID_UNICAST_LOCATOR(PIDPacketBase):
    name = "PID_UNICAST_LOCATOR"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("locator", "", LocatorPacket),
    ]


class PID_MULTICAST_LOCATOR(PIDPacketBase):
    name = "PID_MULTICAST_LOCATOR"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        StrLenField(
            "parameterData",
            "",
            length_from=lambda x: x.parameterLength,
            max_length=STR_MAX_LEN,
        ),
    ]


class PID_MULTICAST_IPADDRESS(PIDPacketBase):
    name = "PID_MULTICAST_IPADDRESS"


class PID_DEFAULT_UNICAST_LOCATOR(PIDPacketBase):
    name = "PID_DEFAULT_UNICAST_LOCATOR"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("locator", "", LocatorPacket),
    ]


class PID_DEFAULT_MULTICAST_LOCATOR(PIDPacketBase):
    name = "PID_DEFAULT_MULTICAST_LOCATOR"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("locator", "", LocatorPacket),
    ]


class PID_TRANSPORT_PRIORITY(PIDPacketBase):
    name = "PID_TRANSPORT_PRIORITY"


class PID_METATRAFFIC_UNICAST_LOCATOR(PIDPacketBase):
    name = "PID_METATRAFFIC_UNICAST_LOCATOR"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("locator", "", LocatorPacket),
    ]


class PID_METATRAFFIC_MULTICAST_LOCATOR(PIDPacketBase):
    name = "PID_METATRAFFIC_MULTICAST_LOCATOR"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("locator", "", LocatorPacket),
    ]


class PID_DEFAULT_UNICAST_IPADDRESS(PIDPacketBase):
    name = "PID_DEFAULT_UNICAST_IPADDRESS"


class PID_DEFAULT_UNICAST_PORT(PIDPacketBase):
    name = "PID_DEFAULT_UNICAST_PORT"


class PID_METATRAFFIC_UNICAST_IPADDRESS(PIDPacketBase):
    name = "PID_METATRAFFIC_UNICAST_IPADDRESS"


class PID_METATRAFFIC_UNICAST_PORT(PIDPacketBase):
    name = "PID_METATRAFFIC_UNICAST_PORT"


class PID_METATRAFFIC_MULTICAST_IPADDRESS(PIDPacketBase):
    name = "PID_METATRAFFIC_MULTICAST_IPADDRESS"


class PID_METATRAFFIC_MULTICAST_PORT(PIDPacketBase):
    name = "PID_METATRAFFIC_MULTICAST_PORT"


class PID_EXPECTS_INLINE_QOS(PIDPacketBase):
    name = "PID_EXPECTS_INLINE_QOS"


class PID_PARTICIPANT_MANUAL_LIVELINESS_COUNT(PIDPacketBase):
    name = "PID_PARTICIPANT_MANUAL_LIVELINESS_COUNT"


class PID_PARTICIPANT_BUILTIN_ENDPOINTS(PIDPacketBase):
    name = "PID_PARTICIPANT_BUILTIN_ENDPOINTS"


class PID_PARTICIPANT_LEASE_DURATION(PIDPacketBase):
    name = "PID_PARTICIPANT_LEASE_DURATION"


class PID_CONTENT_FILTER_PROPERTY(PIDPacketBase):
    name = "PID_CONTENT_FILTER_PROPERTY"


class PID_PARTICIPANT_GUID(PIDPacketBase):
    name = "PID_PARTICIPANT_GUID"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("guid", "", GUIDPacket),
    ]


class PID_ENDPOINT_GUID(PIDPacketBase):
    name = "PID_ENDPOINT_GUID"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("guid", "", GUIDPacket),
    ]


class PID_GROUP_GUID(PIDPacketBase):
    name = "PID_GROUP_GUID"


class PID_GROUP_ENTITYID(PIDPacketBase):
    name = "PID_GROUP_ENTITYID"


class PID_BUILTIN_ENDPOINT_SET(PIDPacketBase):
    name = "PID_BUILTIN_ENDPOINT_SET"


class PID_PROPERTY_LIST(PIDPacketBase):
    name = "PID_PROPERTY_LIST"


class PID_TYPE_MAX_SIZE_SERIALIZED(PIDPacketBase):
    name = "PID_TYPE_MAX_SIZE_SERIALIZED"


class PID_ENTITY_NAME(PIDPacketBase):
    name = "PID_ENTITY_NAME"


class PID_KEY_HASH(PIDPacketBase):
    name = "PID_KEY_HASH"


class PID_STATUS_INFO(PIDPacketBase):
    name = "PID_STATUS_INFO"


class PID_BUILTIN_ENDPOINT_QOS(PIDPacketBase):
    name = "PID_BUILTIN_ENDPOINT_QOS"


class PID_DOMAIN_TAG(PIDPacketBase):
    name = "PID_DOMAIN_TAG"


class PID_DOMAIN_ID(PIDPacketBase):
    name = "PID_DOMAIN_ID"


class PID_UNKNOWN(PIDPacketBase):
    name = "PID_UNKNOWN"


class PID_PRODUCT_VERSION(PIDPacketBase):
    name = "PID_PRODUCT_VERSION"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("productVersion", "", ProductVersionPacket),
    ]


class PID_PLUGIN_PROMISCUITY_KIND(PIDPacketBase):
    name = "PID_PLUGIN_PROMISCUITY_KIND"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        EField(
            XIntField("promiscuityKind", 0x0),
            endianness=FORMAT_LE,
            endianness_from=None
        )
    ]


class PID_RTI_DOMAIN_ID(PIDPacketBase):
    name = "PID_RTI_DOMAIN_ID"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        EField(
            IntField("domainId", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        )
    ]


class PID_TRANSPORT_INFO_LIST(PIDPacketBase):
    name = "PID_TRANSPORT_INFO_LIST"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        XStrFixedLenField("padding", "", 4),
        EField(
            PacketListField(
                "transportInfo", [],
                TransportInfoPacket,
                length_from=lambda p: p.parameterLength - 4)
        )
    ]


class PID_REACHABILITY_LEASE_DURATION(PIDPacketBase):
    name = "PID_REACHABILITY_LEASE_DURATION"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        PacketField("lease_duration", "", LeaseDurationPacket),
    ]


class PID_VENDOR_BUILTIN_ENDPOINT_SET(PIDPacketBase):
    name = "PID_VENDOR_BUILTIN_ENDPOINT_SET"
    fields_desc = [
        EField(
            ParameterIdField("parameterId", 0),
            endianness=FORMAT_LE,
            endianness_from=None,
        ),
        EField(
            ShortField("parameterLength", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        ),
        EField(
            XIntField("flags", 0),
            endianness=FORMAT_LE,
            endianness_from=None
        )
    ]


_RTPSParameterIdTypes = {
    0x0000: PID_PAD,
    # 0x0001: PID_SENTINEL,
    0x0002: PID_PARTICIPANT_LEASE_DURATION,
    0x0004: PID_TIME_BASED_FILTER,
    0x0005: PID_TOPIC_NAME,
    0x0006: PID_OWNERSHIP_STRENGTH,
    0x0007: PID_TYPE_NAME,
    0x000B: PID_METATRAFFIC_MULTICAST_IPADDRESS,
    0x000C: PID_DEFAULT_UNICAST_IPADDRESS,
    0x000D: PID_METATRAFFIC_UNICAST_PORT,
    0x000E: PID_DEFAULT_UNICAST_PORT,
    0x000F: PID_DOMAIN_ID,
    0x0011: PID_MULTICAST_IPADDRESS,
    0x0015: PID_PROTOCOL_VERSION,
    0x0016: PID_VENDOR_ID,
    0x001A: PID_RELIABILITY,
    0x001B: PID_LIVELINESS,
    0x001D: PID_DURABILITY,
    0x001E: PID_DURABILITY_SERVICE,
    0x001F: PID_OWNERSHIP,
    0x0021: PID_PRESENTATION,
    0x0023: PID_DEADLINE,
    0x0025: PID_DESTINATION_ORDER,
    0x0027: PID_LATENCY_BUDGET,
    0x0029: PID_PARTITION,
    0x002B: PID_LIFESPAN,
    0x002C: PID_USER_DATA,
    0x002D: PID_GROUP_DATA,
    0x002E: PID_TOPIC_DATA,
    0x002F: PID_UNICAST_LOCATOR,
    0x0030: PID_MULTICAST_LOCATOR,
    0x0031: PID_DEFAULT_UNICAST_LOCATOR,
    0x0032: PID_METATRAFFIC_UNICAST_LOCATOR,
    0x0033: PID_METATRAFFIC_MULTICAST_LOCATOR,
    0x0034: PID_PARTICIPANT_MANUAL_LIVELINESS_COUNT,
    0x0035: PID_CONTENT_FILTER_PROPERTY,
    0x0040: PID_HISTORY,
    0x0041: PID_RESOURCE_LIMITS,
    0x0043: PID_EXPECTS_INLINE_QOS,
    0x0044: PID_PARTICIPANT_BUILTIN_ENDPOINTS,
    0x0045: PID_METATRAFFIC_UNICAST_IPADDRESS,
    0x0046: PID_METATRAFFIC_MULTICAST_PORT,
    0x0048: PID_DEFAULT_MULTICAST_LOCATOR,
    0x0049: PID_TRANSPORT_PRIORITY,
    0x0050: PID_PARTICIPANT_GUID,
    0x0052: PID_GROUP_GUID,
    0x0053: PID_GROUP_ENTITYID,
    0x0058: PID_BUILTIN_ENDPOINT_SET,
    0x0059: PID_PROPERTY_LIST,
    0x005A: PID_ENDPOINT_GUID,
    0x0060: PID_TYPE_MAX_SIZE_SERIALIZED,
    0x0062: PID_ENTITY_NAME,
    0x0070: PID_KEY_HASH,
    0x0071: PID_STATUS_INFO,
    0x0077: PID_BUILTIN_ENDPOINT_QOS,
    0x4014: PID_DOMAIN_TAG,
    0x8000: PID_PRODUCT_VERSION,
    0x8001: PID_PLUGIN_PROMISCUITY_KIND,
    0x800f: PID_RTI_DOMAIN_ID,
    0x8010: PID_TRANSPORT_INFO_LIST,
    0x8016: PID_REACHABILITY_LEASE_DURATION,
    0x8017: PID_VENDOR_BUILTIN_ENDPOINT_SET
}


def get_pid_class(pkt, lst, cur, remain):

    endianness = getattr(pkt, "endianness", None)

    _id = struct.unpack(endianness + "h", remain[0:2])[0]

    if _id == 0x0001:
        return None

    _id = _id & 0xffff

    next_cls = _RTPSParameterIdTypes.get(_id, PID_UNKNOWN)

    next_cls.endianness = endianness

    return next_cls


class ParameterListPacket(EPacket):
    name = "PID list"
    fields_desc = [
        PacketListField("parameterValues", [], next_cls_cb=get_pid_class),
        PacketField("sentinel", "", PID_SENTINEL),
    ]



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Thomas Tannhaeuser <hecke@naberius.de>

# scapy.contrib.description = IEC-60870-5-104 APCI / APDU layer definitions
# scapy.contrib.status = loads

"""
    IEC 60870-5-104
    ~~~~~~~~~~~~~~~

    :description:

        This module provides the IEC 60870-5-104 (common short name: iec104)
        layer, the information objects and related information element
        definitions.

        normative references:
            - IEC 60870-5-4:1994 (atomic base types / data format)
            - IEC 60870-5-101:2003 (information elements (sec. 7.2.6) and
              ASDU definition (sec. 7.3))
            - IEC 60870-5-104:2006 (information element TSC (sec. 8.8, p. 44))

    :TODO:
        - add allowed direction to IO attributes
          (but this could be derived from the name easily <--> )
        - information elements / objects need more testing
          (e.g. on live traffic w comparison against tshark)

    :NOTES:
        - bit and octet numbering is used as in the related standards
          (they usually start with index one instead of zero)
        - some of the information objects are only valid for IEC 60870-5-101 -
          so usually they should never appear on the network as iec101 uses
          serial connections. I added them if decoding of those messages is
          needed cause one goes to implement a iec101<-->iec104 gateway or
          hits such a gateway that acts not standard conform (e.g. by
          forwarding 101 messages to a 104 network)
"""

from scapy.contrib.scada.iec104.iec104_fields import *  # noqa F403,F401
from scapy.contrib.scada.iec104.iec104_information_elements import *  # noqa F403,F401
from scapy.contrib.scada.iec104.iec104_information_objects import *  # noqa F403,F401

from scapy.compat import orb
from scapy.config import conf
from scapy.error import warning, Scapy_Exception
from scapy.fields import ByteField, BitField, ByteEnumField, PacketListField, \
    BitEnumField, XByteField, FieldLenField, LEShortField, BitFieldLenField
from scapy.layers.inet import TCP
from scapy.packet import Raw, Packet, bind_layers

IEC_104_IANA_PORT = 2404

# direction - from the central station to the substation
IEC104_CONTROL_DIRECTION = 0
IEC104_CENTRAL_2_SUB_DIR = IEC104_CONTROL_DIRECTION

# direction - from the substation to the central station
IEC104_MONITOR_DIRECTION = 1
IEC104_SUB_2_CENTRAL_DIR = IEC104_MONITOR_DIRECTION

IEC104_DIRECTIONS = {
    IEC104_MONITOR_DIRECTION: 'monitor direction (sub -> central)',
    IEC104_CONTROL_DIRECTION: 'control direction (central -> sub)',
}

# COT - cause of transmission
IEC104_COT_UNDEFINED = 0
IEC104_COT_CYC = 1
IEC104_COT_BACK = 2
IEC104_COT_SPONT = 3
IEC104_COT_INIT = 4
IEC104_COT_REQ = 5
IEC104_COT_ACT = 6
IEC104_COT_ACTCON = 7
IEC104_COT_DEACT = 8
IEC104_COT_DEACTCON = 9
IEC104_COT_ACTTERM = 10
IEC104_COT_RETREM = 11
IEC104_COT_RETLOC = 12
IEC104_COT_FILE = 13
IEC104_COT_RESERVED_14 = 14
IEC104_COT_RESERVED_15 = 15
IEC104_COT_RESERVED_16 = 16
IEC104_COT_RESERVED_17 = 17
IEC104_COT_RESERVED_18 = 18
IEC104_COT_RESERVED_19 = 19
IEC104_COT_INROGEN = 20
IEC104_COT_INRO1 = 21
IEC104_COT_INRO2 = 22
IEC104_COT_INRO3 = 23
IEC104_COT_INRO4 = 24
IEC104_COT_INRO5 = 25
IEC104_COT_INRO6 = 26
IEC104_COT_INRO7 = 27
IEC104_COT_INRO8 = 28
IEC104_COT_INRO9 = 29
IEC104_COT_INRO10 = 30
IEC104_COT_INRO11 = 31
IEC104_COT_INRO12 = 32
IEC104_COT_INRO13 = 33
IEC104_COT_INRO14 = 34
IEC104_COT_INRO15 = 35
IEC104_COT_INRO16 = 36
IEC104_COT_REQCOGEN = 37
IEC104_COT_REQCO1 = 38
IEC104_COT_REQCO2 = 39
IEC104_COT_REQCO3 = 40
IEC104_COT_REQCO4 = 41
IEC104_COT_RESERVED_42 = 42
IEC104_COT_RESERVED_43 = 43
IEC104_COT_UNKNOWN_TYPE_CODE = 44
IEC104_COT_UNKNOWN_TRANSMIT_REASON = 45
IEC104_COT_UNKNOWN_COMMON_ADDRESS_OF_ASDU = 46
IEC104_COT_UNKNOWN_ADDRESS_OF_INFORMATION_OBJECT = 47
IEC104_COT_PRIVATE_48 = 48
IEC104_COT_PRIVATE_49 = 49
IEC104_COT_PRIVATE_50 = 50
IEC104_COT_PRIVATE_51 = 51
IEC104_COT_PRIVATE_52 = 52
IEC104_COT_PRIVATE_53 = 53
IEC104_COT_PRIVATE_54 = 54
IEC104_COT_PRIVATE_55 = 55
IEC104_COT_PRIVATE_56 = 56
IEC104_COT_PRIVATE_57 = 57
IEC104_COT_PRIVATE_58 = 58
IEC104_COT_PRIVATE_59 = 59
IEC104_COT_PRIVATE_60 = 60
IEC104_COT_PRIVATE_61 = 61
IEC104_COT_PRIVATE_62 = 62
IEC104_COT_PRIVATE_63 = 63

CAUSE_OF_TRANSMISSIONS = {
    IEC104_COT_UNDEFINED: 'undefined',
    IEC104_COT_CYC: 'cyclic (per/cyc)',
    IEC104_COT_BACK: 'background (back)',
    IEC104_COT_SPONT: 'spontaneous (spont)',
    IEC104_COT_INIT: 'initialized (init)',
    IEC104_COT_REQ: 'request (req)',
    IEC104_COT_ACT: 'activation (act)',
    IEC104_COT_ACTCON: 'activation confirmed (actcon)',
    IEC104_COT_DEACT: 'activation canceled (deact)',
    IEC104_COT_DEACTCON: 'activation cancellation confirmed (deactcon)',
    IEC104_COT_ACTTERM: 'activation finished (actterm)',
    IEC104_COT_RETREM: 'feedback caused by remote command (retrem)',
    IEC104_COT_RETLOC: 'feedback caused by local command (retloc)',
    IEC104_COT_FILE: 'file transfer (file)',
    IEC104_COT_RESERVED_14: 'reserved_14',
    IEC104_COT_RESERVED_15: 'reserved_15',
    IEC104_COT_RESERVED_16: 'reserved_16',
    IEC104_COT_RESERVED_17: 'reserved_17',
    IEC104_COT_RESERVED_18: 'reserved_18',
    IEC104_COT_RESERVED_19: 'reserved_19',
    IEC104_COT_INROGEN: 'queried by station (inrogen)',
    IEC104_COT_INRO1: 'queried by query to group 1 (inro1)',
    IEC104_COT_INRO2: 'queried by query to group 2 (inro2)',
    IEC104_COT_INRO3: 'queried by query to group 3 (inro3)',
    IEC104_COT_INRO4: 'queried by query to group 4 (inro4)',
    IEC104_COT_INRO5: 'queried by query to group 5 (inro5)',
    IEC104_COT_INRO6: 'queried by query to group 6 (inro6)',
    IEC104_COT_INRO7: 'queried by query to group 7 (inro7)',
    IEC104_COT_INRO8: 'queried by query to group 8 (inro8)',
    IEC104_COT_INRO9: 'queried by query to group 9 (inro9)',
    IEC104_COT_INRO10: 'queried by query to group 10 (inro10)',
    IEC104_COT_INRO11: 'queried by query to group 11 (inro11)',
    IEC104_COT_INRO12: 'queried by query to group 12 (inro12)',
    IEC104_COT_INRO13: 'queried by query to group 13 (inro13)',
    IEC104_COT_INRO14: 'queried by query to group 14 (inro14)',
    IEC104_COT_INRO15: 'queried by query to group 15 (inro15)',
    IEC104_COT_INRO16: 'queried by query to group 16 (inro16)',
    IEC104_COT_REQCOGEN: 'queried by counter general interrogation (reqcogen)',
    IEC104_COT_REQCO1: 'queried by query to counter group 1 (reqco1)',
    IEC104_COT_REQCO2: 'queried by query to counter group 2 (reqco2)',
    IEC104_COT_REQCO3: 'queried by query to counter group 3 (reqco3)',
    IEC104_COT_REQCO4: 'queried by query to counter group 4 (reqco4)',
    IEC104_COT_RESERVED_42: 'reserved_42',
    IEC104_COT_RESERVED_43: 'reserved_43',
    IEC104_COT_UNKNOWN_TYPE_CODE: 'unknown type code',
    IEC104_COT_UNKNOWN_TRANSMIT_REASON: 'unknown transmit reason',
    IEC104_COT_UNKNOWN_COMMON_ADDRESS_OF_ASDU:
        'unknown common address of ASDU',
    IEC104_COT_UNKNOWN_ADDRESS_OF_INFORMATION_OBJECT:
        'unknown address of information object',
    IEC104_COT_PRIVATE_48: 'private_48',
    IEC104_COT_PRIVATE_49: 'private_49',
    IEC104_COT_PRIVATE_50: 'private_50',
    IEC104_COT_PRIVATE_51: 'private_51',
    IEC104_COT_PRIVATE_52: 'private_52',
    IEC104_COT_PRIVATE_53: 'private_53',
    IEC104_COT_PRIVATE_54: 'private_54',
    IEC104_COT_PRIVATE_55: 'private_55',
    IEC104_COT_PRIVATE_56: 'private_56',
    IEC104_COT_PRIVATE_57: 'private_57',
    IEC104_COT_PRIVATE_58: 'private_58',
    IEC104_COT_PRIVATE_59: 'private_59',
    IEC104_COT_PRIVATE_60: 'private_60',
    IEC104_COT_PRIVATE_61: 'private_61',
    IEC104_COT_PRIVATE_62: 'private_62',
    IEC104_COT_PRIVATE_63: 'private_63'
}

IEC104_APDU_TYPE_UNKNOWN = 0x00
IEC104_APDU_TYPE_I_SEQ_IOA = 0x01
IEC104_APDU_TYPE_I_SINGLE_IOA = 0x02
IEC104_APDU_TYPE_U = 0x03
IEC104_APDU_TYPE_S = 0x04


def _iec104_apci_type_from_packet(data):
    """
    the type of the message is encoded in octet 1..4

                 oct 1, bit 1   2       oct 3, bit 1
    I Message               0  1|0                 0
    S Message               1   0                  0
    U Message               1   1                  0


    see EN 60870-5-104:2006, sec. 5 (p. 13, fig. 6,7,8)
    """

    oct_1 = orb(data[2])
    oct_3 = orb(data[4])

    oct_1_bit_1 = bool(oct_1 & 1)
    oct_1_bit_2 = bool(oct_1 & 2)
    oct_3_bit_1 = bool(oct_3 & 1)

    if oct_1_bit_1 is False and oct_3_bit_1 is False:
        if len(data) < 8:
            return IEC104_APDU_TYPE_UNKNOWN

        is_seq_ioa = ((orb(data[7]) & 0x80) == 0x80)

        if is_seq_ioa:
            return IEC104_APDU_TYPE_I_SEQ_IOA
        else:
            return IEC104_APDU_TYPE_I_SINGLE_IOA

    if oct_1_bit_1 and oct_1_bit_2 is False and oct_3_bit_1 is False:
        return IEC104_APDU_TYPE_S

    if oct_1_bit_1 and oct_1_bit_2 and oct_3_bit_1 is False:
        return IEC104_APDU_TYPE_U

    return IEC104_APDU_TYPE_UNKNOWN


class IEC104_APDU(Packet):
    """
    basic Application Protocol Data Unit definition used by S/U/I messages
    """

    def guess_payload_class(self, payload):

        payload_len = len(payload)

        if payload_len < 6:
            return self.default_payload_class(payload)

        if orb(payload[0]) != 0x68:
            self.default_payload_class(payload)

        # the length field contains the number of bytes starting from the
        # first control octet
        apdu_length = 2 + orb(payload[1])

        if payload_len < apdu_length:
            warning(
                'invalid len of APDU. given len: {} available len: {}'.format(
                    apdu_length, payload_len))
            return self.default_payload_class(payload)

        apdu_type = _iec104_apci_type_from_packet(payload)

        return IEC104_APDU_CLASSES.get(apdu_type,
                                       self.default_payload_class(payload))

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        """
        detect type of the message by checking packet data
        :param _pkt: raw bytes of the packet layer data to be checked
        :param args: unused
        :param kargs: unused
        :return: class of the detected message type
        """

        if _iec104_is_i_apdu_seq_ioa(_pkt):
            return IEC104_I_Message_SeqIOA

        if _iec104_is_i_apdu_single_ioa(_pkt):
            return IEC104_I_Message_SingleIOA

        if _iec104_is_u_apdu(_pkt):
            return IEC104_U_Message

        if _iec104_is_s_apdu(_pkt):
            return IEC104_S_Message

        return Raw


class IEC104_S_Message(IEC104_APDU):
    """
    message used for ack of received I-messages
    """
    name = 'IEC-104 S APDU'

    fields_desc = [

        XByteField('start', 0x68),
        ByteField("apdu_length", 4),

        ByteField('octet_1', 0x01),
        ByteField('octet_2', 0),
        IEC104SequenceNumber('rx_seq_num', 0),
    ]


class IEC104_U_Message(IEC104_APDU):
    """
    message used for connection tx control (start/stop)  and monitoring (test)
    """
    name = 'IEC-104 U APDU'

    fields_desc = [

        XByteField('start', 0x68),
        ByteField("apdu_length", 4),

        BitField('testfr_con', 0, 1),
        BitField('testfr_act', 0, 1),
        BitField('stopdt_con', 0, 1),
        BitField('stopdt_act', 0, 1),
        BitField('startdt_con', 0, 1),
        BitField('startdt_act', 0, 1),
        BitField('octet_1_1_2', 3, 2),

        ByteField('octet_2', 0),
        ByteField('octet_3', 0),
        ByteField('octet_4', 0)
    ]


def _i_msg_io_dispatcher_sequence(pkt, next_layer_data):
    """
    get the type id and return the matching ASDU instance
    """
    next_layer_class_type = IEC104_IO_CLASSES.get(pkt.type_id, conf.raw_layer)

    return next_layer_class_type(next_layer_data)


def _i_msg_io_dispatcher_single(pkt, next_layer_data):
    """
    get the type id and return the matching ASDU instance
    (information object address + regular ASDU information object fields)
    """
    next_layer_class_type = IEC104_IO_WITH_IOA_CLASSES.get(pkt.type_id,
                                                           conf.raw_layer)

    return next_layer_class_type(next_layer_data)


class IEC104ASDUPacketListField(PacketListField):
    """
    used to add a list of information objects to an I-message
    """
    def m2i(self, pkt, m):
        """
        add calling layer instance to the cls()-signature
        :param pkt: calling layer instance
        :param m: raw data forming the next layer
        :return: instance of the class representing the next layer
        """
        return self.cls(pkt, m)


class IEC104_I_Message_StructureException(Scapy_Exception):
    """
    Exception raised if payload is not of type Information Object
    """
    pass


class IEC104_I_Message(IEC104_APDU):
    """
    message used for transmitting data (APDU - Application Protocol Data Unit)

    APDU: MAGIC + APCI + ASDU
    MAGIC: 0x68
    APCI : Control Information (rx/tx seq/ack numbers)
    ASDU : Application Service Data Unit - information object related data

    see EN 60870-5-104:2006, sec. 5 (p. 12)
    """
    name = 'IEC-104 I APDU'

    IEC_104_MAGIC = 0x68  # dec -> 104

    SQ_FLAG_SINGLE = 0
    SQ_FLAG_SEQUENCE = 1

    SQ_FLAGS = {
        SQ_FLAG_SINGLE: 'single',
        SQ_FLAG_SEQUENCE: 'sequence'
    }

    TEST_DISABLED = 0
    TEST_ENABLED = 1

    TEST_FLAGS = {
        TEST_DISABLED: 'disabled',
        TEST_ENABLED: 'enabled'
    }

    ACK_POSITIVE = 0
    ACK_NEGATIVE = 1

    ACK_FLAGS = {
        ACK_POSITIVE: 'positive',
        ACK_NEGATIVE: 'negative'
    }

    fields_desc = []

    def __init__(self, _pkt=b"", post_transform=None, _internal=0,
                 _underlayer=None, **fields):

        super(IEC104_I_Message, self).__init__(_pkt=_pkt,
                                               post_transform=post_transform,
                                               _internal=_internal,
                                               _underlayer=_underlayer,
                                               **fields)

        if 'io' in fields and fields['io']:
            self._information_object_update(fields['io'])

    def _information_object_update(self, io_instances):
        """
        set the type_id in the ASDU header based on the given information
        object (io) and check for valid structure
        :param io_instances: information object
        """

        if not isinstance(io_instances, list):
            io_instances = [io_instances]

        first_io = io_instances[0]
        first_io_class = first_io.__class__

        if not issubclass(first_io_class, IEC104_IO_Packet):
            raise IEC104_I_Message_StructureException(
                'information object payload must be a subclass of '
                'IEC104_IO_Packet')

        self.type_id = first_io.iec104_io_type_id()

        # ensure all io elements within the ASDU share the same class type
        for io_inst in io_instances[1:]:
            if io_inst.__class__ != first_io_class:
                raise IEC104_I_Message_StructureException(
                    'each information object within the ASDU must be of '
                    'the same class type (first io: {}, '
                    'current io: {})'.format(first_io_class._name,
                                             io_inst._name))


class IEC104_I_Message_SeqIOA(IEC104_I_Message):
    """
    all information objects share a base information object address field

    sq = 1, see EN 60870-5-101:2003, sec. 7.2.2.1 (p. 33)
    """
    name = 'IEC-104 I APDU (Seq IOA)'

    fields_desc = [
        # APCI
        XByteField('start', IEC104_I_Message.IEC_104_MAGIC),
        FieldLenField("apdu_length", None, fmt="!B", length_of='io',
                      adjust=lambda pkt, x: x + 13),

        IEC104SequenceNumber('tx_seq_num', 0),
        IEC104SequenceNumber('rx_seq_num', 0),

        # ASDU
        ByteEnumField('type_id', 0, IEC104_IO_NAMES),

        BitEnumField('sq', IEC104_I_Message.SQ_FLAG_SEQUENCE, 1,
                     IEC104_I_Message.SQ_FLAGS),
        BitFieldLenField('num_io', None, 7, count_of='io'),

        BitEnumField('test', 0, 1, IEC104_I_Message.TEST_FLAGS),
        BitEnumField('ack', 0, 1, IEC104_I_Message.ACK_FLAGS),
        BitEnumField('cot', 0, 6, CAUSE_OF_TRANSMISSIONS),

        ByteField('origin_address', 0),

        LEShortField('common_asdu_address', 0),

        LEThreeBytesField('information_object_address', 0),

        IEC104ASDUPacketListField('io',
                                  conf.raw_layer(),
                                  _i_msg_io_dispatcher_sequence,
                                  length_from=lambda pkt: pkt.apdu_length - 13)
    ]

    def post_dissect(self, s):
        if self.type_id == IEC104_IO_ID_C_RD_NA_1:

            # IEC104_IO_ID_C_RD_NA_1 has no payload. we will add the layer
            # manually to the stack right now. we do this num_io times
            # as - even if it makes no sense - someone could decide
            # to add more than one read commands in a sequence...
            setattr(self, 'io', [IEC104_IO_C_RD_NA_1()] * self.num_io)

        return s


class IEC104_I_Message_SingleIOA(IEC104_I_Message):
    """
    every information object contains an individual information object
    address field

    sq = 0, see EN 60870-5-101:2003, sec. 7.2.2.1 (p. 33)
    """
    name = 'IEC-104 I APDU (single IOA)'

    fields_desc = [
        # APCI
        XByteField('start', IEC104_I_Message.IEC_104_MAGIC),
        FieldLenField("apdu_length", None, fmt="!B", length_of='io',
                      adjust=lambda pkt, x: x + 10),

        IEC104SequenceNumber('tx_seq_num', 0),
        IEC104SequenceNumber('rx_seq_num', 0),

        # ASDU
        ByteEnumField('type_id', 0, IEC104_IO_NAMES),

        BitEnumField('sq', IEC104_I_Message.SQ_FLAG_SINGLE, 1,
                     IEC104_I_Message.SQ_FLAGS),
        BitFieldLenField('num_io', None, 7, count_of='io'),

        BitEnumField('test', 0, 1, IEC104_I_Message.TEST_FLAGS),
        BitEnumField('ack', 0, 1, IEC104_I_Message.ACK_FLAGS),
        BitEnumField('cot', 0, 6, CAUSE_OF_TRANSMISSIONS),

        ByteField('origin_address', 0),

        LEShortField('common_asdu_address', 0),

        IEC104ASDUPacketListField('io',
                                  conf.raw_layer(),
                                  _i_msg_io_dispatcher_single,
                                  length_from=lambda pkt: pkt.apdu_length - 10)
    ]


IEC104_APDU_CLASSES = {
    IEC104_APDU_TYPE_UNKNOWN: conf.raw_layer,
    IEC104_APDU_TYPE_I_SEQ_IOA: IEC104_I_Message_SeqIOA,
    IEC104_APDU_TYPE_I_SINGLE_IOA: IEC104_I_Message_SingleIOA,
    IEC104_APDU_TYPE_U: IEC104_U_Message,
    IEC104_APDU_TYPE_S: IEC104_S_Message
}


def _iec104_is_i_apdu_seq_ioa(payload):
    len_payload = len(payload)
    if len_payload < 6:
        return False

    if orb(payload[0]) != 0x68 or (
            orb(payload[1]) + 2) > len_payload or len_payload < 8:
        return False

    return IEC104_APDU_TYPE_I_SEQ_IOA == _iec104_apci_type_from_packet(payload)


def _iec104_is_i_apdu_single_ioa(payload):
    len_payload = len(payload)
    if len_payload < 6:
        return False

    if orb(payload[0]) != 0x68 or (
            orb(payload[1]) + 2) > len_payload or len_payload < 8:
        return False

    return IEC104_APDU_TYPE_I_SINGLE_IOA == _iec104_apci_type_from_packet(
        payload)


def _iec104_is_u_apdu(payload):
    if len(payload) < 6:
        return False

    if orb(payload[0]) != 0x68 or orb(payload[1]) != 4:
        return False

    return IEC104_APDU_TYPE_U == _iec104_apci_type_from_packet(payload)


def _iec104_is_s_apdu(payload):
    if len(payload) < 6:
        return False

    if orb(payload[0]) != 0x68 or orb(payload[1]) != 4:
        return False

    return IEC104_APDU_TYPE_S == _iec104_apci_type_from_packet(payload)


def iec104_decode(payload):
    """
    can be used to dissect payload of a TCP connection
    :param payload: the application layer data (IEC104-APDU(s))
    :return: iec104 (I/U/S) message instance, conf.raw_layer() if unknown
    """

    if _iec104_is_i_apdu_seq_ioa(payload):
        return IEC104_I_Message_SeqIOA(payload)
    elif _iec104_is_i_apdu_single_ioa(payload):
        return IEC104_I_Message_SingleIOA(payload)
    elif _iec104_is_s_apdu(payload):
        return IEC104_S_Message(payload)
    elif _iec104_is_u_apdu(payload):
        return IEC104_U_Message(payload)
    else:
        return conf.raw_layer(payload)


bind_layers(TCP, IEC104_APDU, sport=IEC_104_IANA_PORT)
bind_layers(TCP, IEC104_APDU, dport=IEC_104_IANA_PORT)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
Secure Shell (SSH) Transport Layer Protocol

RFC 4250, 4251, 4252, 4253 and 4254
"""

from scapy.config import conf
from scapy.compat import plain_str
from scapy.fields import (
    BitLenField,
    ByteField,
    ByteEnumField,
    IntEnumField,
    IntField,
    PacketField,
    PacketListField,
    PacketLenField,
    FieldLenField,
    FieldListField,
    StrLenField,
    StrFixedLenField,
    StrNullField,
    YesNoByteField,
)
from scapy.packet import Packet, bind_bottom_up, bind_layers

from scapy.layers.inet import TCP


class StrCRLFField(StrNullField):
    DELIMITER = b"\r\n"


class _SSHHeaderField(FieldListField):
    def getfield(self, pkt, s):
        val = []
        while s:
            s, v = self.field.getfield(pkt, s)
            val.append(v)
            if v[:4] == b"SSH-":
                return s, val
        return s, val


# RFC 4251 - SSH Architecture
# This RFC defines some types

# RFC 4251 - sect 5


class _ComaStrField(StrLenField):
    islist = 1

    def m2i(self, pkt, x):
        return super(_ComaStrField, self).m2i(pkt, x).split(b",")

    def i2m(self, pkt, x):
        return super(_ComaStrField, self).i2m(pkt, b",".join(x))


class SSHString(Packet):
    fields_desc = [
        FieldLenField("length", None, length_of="value", fmt="!I"),
        StrLenField("value", 0, length_from=lambda pkt: pkt.length),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


class SSHPacketStringField(PacketField):
    __slots__ = ["sub_cls"]

    def __init__(self, name, sub_cls):
        self.sub_cls = sub_cls
        super(SSHPacketStringField, self).__init__(name, SSHString(), SSHString)

    def m2i(self, pkt, x):
        x = super(SSHPacketStringField, self).m2i(pkt, x)
        x.value = self.sub_cls(x.value)
        return x


class NameList(Packet):
    fields_desc = [
        FieldLenField("length", None, length_of="names", fmt="!I"),
        _ComaStrField("names", [], length_from=lambda pkt: pkt.length),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


class Mpint(Packet):
    fields_desc = [
        FieldLenField("length", None, length_of="value", fmt="!I"),
        BitLenField("value", 0, length_from=lambda pkt: pkt.length * 8),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# RFC4250 - sect 4.1.2

_SSH_message_numbers = {
    # RFC4253 - SSH-TRANS
    1: "SSH_MSG_DISCONNECT",
    2: "SSH_MSG_IGNORE",
    3: "SSH_MSG_UNIMPLEMENTED",
    4: "SSH_MSG_DEBUG",
    5: "SSH_MSG_SERVICE_REQUEST",
    6: "SSH_MSG_SERVICE_ACCEPT",
    7: "SSH_MSG_EXT_INFO",  # RFC 8308
    8: "SSH_MSG_NEWCOMPRESS",
    20: "SSH_MSG_KEXINIT",
    21: "SSH_MSG_NEWKEYS",
    # Errata 152 of RFC4253
    30: "SSH_MSG_KEXDH_INIT",
    31: "SSH_MSG_KEXDH_REPLY",
    # RFC4252 - SSH-USERAUTH
    50: "SSH_MSG_USERAUTH_REQUEST",
    51: "SSH_MSG_USERAUTH_FAILURE",
    52: "SSH_MSG_USERAUTH_SUCCESS",
    53: "SSH_MSG_USERAUTH_BANNER",
    # RFC4254 - SSH-CONNECT
    80: "SSH_MSG_GLOBAL_REQUEST",
    81: "SSH_MSG_REQUEST_SUCCESS",
    82: "SSH_MSG_REQUEST_FAILURE",
    90: "SSH_MSG_CHANNEL_OPEN",
    91: "SSH_MSG_CHANNEL_OPEN_CONFIRMATION",
    92: "SSH_MSG_CHANNEL_OPEN_FAILURE",
    93: "SSH_MSG_CHANNEL_WINDOW_ADJUST",
    94: "SSH_MSG_CHANNEL_DATA",
    95: "SSH_MSG_CHANNEL_EXTENDED_DATA",
    96: "SSH_MSG_CHANNEL_EOF",
    97: "SSH_MSG_CHANNEL_CLOSE",
    98: "SSH_MSG_CHANNEL_REQUEST",
    99: "SSH_MSG_CHANNEL_SUCCESS",
    100: "SSH_MSG_CHANNEL_FAILURE",
}

# RFC4253 - sect 6

_SSH_messages = {}


def _SSHPayload(x, **kwargs):
    return _SSH_messages.get(x and x[0], conf.raw_layer)(x)


class SSH(Packet):
    name = "SSH - Binary Packet"
    fields_desc = [
        IntField("packet_length", None),
        ByteField("padding_length", None),
        PacketLenField(
            "pay",
            None,
            _SSHPayload,
            length_from=lambda pkt: pkt.packet_length - pkt.padding_length - 1,
        ),
        StrLenField("random_padding", b"", length_from=lambda pkt: pkt.padding_length),
        # StrField("mac", b""),
    ]

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and len(_pkt) >= 4 and _pkt[:4] == b"SSH-":
            return SSHVersionExchange
        return cls

    def mysummary(self):
        if self.pay:
            if isinstance(self.pay, conf.raw_layer):
                return "SSH type " + str(self.pay.load[0]), [TCP, SSH]
            return "SSH " + self.pay.sprintf("%type%"), [TCP, SSH]
        return "SSH", [TCP, SSH]


# RFC4253 - sect 4.2


class SSHVersionExchange(Packet):
    name = "SSH - Protocol Version Exchange"
    fields_desc = [
        _SSHHeaderField(
            "lines",
            [],
            StrCRLFField("", b""),
        )
    ]

    def mysummary(self):
        return "SSH - Version Exchange %s" % plain_str(self.lines[-1]), [TCP]


# RFC4253 - sect 6.6

_SSH_certificates = {}
_SSH_publickeys = {}
_SSH_signatures = {}


class _SSHCertificate(PacketField):
    def m2i(self, pkt, x):
        return _SSH_certificates.get(pkt.format_identifier.value, self.cls)(x)


class _SSHPublicKey(PacketField):
    def m2i(self, pkt, x):
        return _SSH_publickeys.get(pkt.format_identifier.value, self.cls)(x)


class _SSHSignature(PacketField):
    def m2i(self, pkt, x):
        return _SSH_signatures.get(pkt.format_identifier.value, self.cls)(x)


class SSHCertificate(Packet):
    fields_desc = [
        PacketField("format_identifier", SSHString(), SSHString),
        _SSHCertificate("data", None, conf.raw_layer),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


class SSHPublicKey(Packet):
    fields_desc = [
        PacketField("format_identifier", SSHString(), SSHString),
        _SSHPublicKey("data", None, conf.raw_layer),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


class SSHSignature(Packet):
    fields_desc = [
        PacketField("format_identifier", SSHString(), SSHString),
        _SSHSignature("data", None, conf.raw_layer),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# RFC4253 - sect 7.1


class SSHKexInit(Packet):
    fields_desc = [
        ByteEnumField("type", 20, _SSH_message_numbers),
        StrFixedLenField("cookie", b"", length=16),
        PacketField("kex_algorithms", NameList(), NameList),
        PacketField("server_host_key_algorithms", NameList(), NameList),
        PacketField("encryption_algorithms_client_to_server", NameList(), NameList),
        PacketField("encryption_algorithms_server_to_client", NameList(), NameList),
        PacketField("mac_algorithms_client_to_server", NameList(), NameList),
        PacketField("mac_algorithms_server_to_client", NameList(), NameList),
        PacketField("compression_algorithms_client_to_server", NameList(), NameList),
        PacketField("compression_algorithms_server_to_client", NameList(), NameList),
        PacketField("languages_client_to_server", NameList(), NameList),
        PacketField("languages_server_to_client", NameList(), NameList),
        YesNoByteField("first_kex_packet_follows", 0),
        IntField("reserved", 0),
    ]


_SSH_messages[20] = SSHKexInit

# RFC4253 - sect 7.3


class SSHNewKeys(Packet):
    fields_desc = [
        ByteEnumField("type", 21, _SSH_message_numbers),
    ]


_SSH_messages[21] = SSHNewKeys


# RFC4253 - sect 8


class SSHKexDHInit(Packet):
    fields_desc = [
        ByteEnumField("type", 30, _SSH_message_numbers),
        PacketField("e", Mpint(), Mpint),
    ]


_SSH_messages[30] = SSHKexDHInit


class SSHKexDHReply(Packet):
    fields_desc = [
        ByteEnumField("type", 31, _SSH_message_numbers),
        SSHPacketStringField("K_S", SSHPublicKey),
        PacketField("f", Mpint(), Mpint),
        SSHPacketStringField("H_hash", SSHSignature),
    ]


_SSH_messages[31] = SSHKexDHReply

# RFC4253 - sect 10


class SSHServiceRequest(Packet):
    fields_desc = [
        ByteEnumField("type", 5, _SSH_message_numbers),
        PacketField("service_name", SSHString(), SSHString),
    ]


_SSH_messages[5] = SSHServiceRequest


class SSHServiceAccept(Packet):
    fields_desc = [
        ByteEnumField("type", 6, _SSH_message_numbers),
        PacketField("service_name", SSHString(), SSHString),
    ]


_SSH_messages[6] = SSHServiceAccept

# RFC4253 - sect 11.1


class SSHDisconnect(Packet):
    fields_desc = [
        ByteEnumField("type", 1, _SSH_message_numbers),
        IntEnumField(
            "reason_code",
            0,
            {
                1: "SSH_DISCONNECT_HOST_NOT_ALLOWED_TO_CONNECT",
                2: "SSH_DISCONNECT_PROTOCOL_ERROR",
                3: "SSH_DISCONNECT_KEY_EXCHANGE_FAILED",
                4: "SSH_DISCONNECT_RESERVED",
                5: "SSH_DISCONNECT_MAC_ERROR",
                6: "SSH_DISCONNECT_COMPRESSION_ERROR",
                7: "SSH_DISCONNECT_SERVICE_NOT_AVAILABLE",
                8: "SSH_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED",
                9: "SSH_DISCONNECT_HOST_KEY_NOT_VERIFIABLE",
                10: "SSH_DISCONNECT_CONNECTION_LOST",
                11: "SSH_DISCONNECT_BY_APPLICATION",
                12: "SSH_DISCONNECT_TOO_MANY_CONNECTIONS",
                13: "SSH_DISCONNECT_AUTH_CANCELLED_BY_USER",
                14: "SSH_DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE",
                15: "SSH_DISCONNECT_ILLEGAL_USER_NAME",
            },
        ),
        PacketField("description", SSHString(), SSHString),
        PacketField("language_tag", SSHString(), SSHString),
    ]


_SSH_messages[1] = SSHDisconnect

# RFC4253 - sect 11.2


class SSHIgnore(Packet):
    fields_desc = [
        ByteEnumField("type", 2, _SSH_message_numbers),
        PacketField("data", SSHString(), SSHString),
    ]


_SSH_messages[2] = SSHIgnore

# RFC4253 - sect 11.3


class SSHServiceDebug(Packet):
    fields_desc = [
        ByteEnumField("type", 4, _SSH_message_numbers),
        YesNoByteField("always_display", 0),
        PacketField("message", SSHString(), SSHString),
        PacketField("language_tag", SSHString(), SSHString),
    ]


_SSH_messages[4] = SSHServiceDebug

# RFC4253 - sect 11.4


class SSHUnimplemented(Packet):
    fields_desc = [
        ByteEnumField("type", 3, _SSH_message_numbers),
        IntField("seq_num", 0),
    ]


_SSH_messages[3] = SSHUnimplemented

# RFC8308 - sect 2.3


class SSHExtension(Packet):
    fields_desc = [
        PacketField("extension_name", SSHString(), SSHString),
        PacketField("extension_value", SSHString(), SSHString),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


class SSHExtInfo(Packet):
    fields_desc = [
        ByteEnumField("type", 7, _SSH_message_numbers),
        FieldLenField("nr_extensions", None, length_of="extensions"),
        PacketListField("extensions", [], SSHExtension),
    ]


_SSH_messages[7] = SSHExtInfo

# RFC8308 - sect 3.2


class SSHNewCompress(Packet):
    fields_desc = [
        ByteEnumField("type", 3, _SSH_message_numbers),
    ]


_SSH_messages[8] = SSHNewCompress

# RFC8709


class SSHPublicKeyEd25519(Packet):
    fields_desc = [
        PacketField("key", SSHString(), SSHString),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


_SSH_publickeys[b"ssh-ed25519"] = SSHPublicKeyEd25519


class SSHPublicKeyEd448(Packet):
    fields_desc = [
        PacketField("key", SSHString(), SSHString),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


_SSH_publickeys[b"ssh-ed448"] = SSHPublicKeyEd448


class SSHSignatureEd25519(Packet):
    fields_desc = [
        PacketField("key", SSHString(), SSHString),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


_SSH_signatures[b"ssh-ed25519"] = SSHSignatureEd25519


class SSHSignatureEd448(Packet):
    fields_desc = [
        PacketField("key", SSHString(), SSHString),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


_SSH_signatures[b"ssh-ed448"] = SSHSignatureEd448

bind_layers(SSH, SSH)

bind_bottom_up(TCP, SSH, sport=22)
bind_layers(TCP, SSH, dport=22)
